import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.utils import tilecode
from vgrid.utils import mercantile
from vgrid.generator.settings import graticule_dggs_to_feature
from vgrid.conversion.dggscompact import tilecodecompact
import re
from vgrid.utils.geometry import (
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)
from math import sqrt


def validate_tilecode_resolution(resolution):
    """
    Validate that Tilecode resolution is in the valid range [0..29] (0=coarsest, 29=finest).

    Args:
        resolution (int): Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [0..29]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 0 or resolution > 29:
        raise ValueError(f"Resolution must be in range [0..29], got {resolution}")

    return resolution


def point2tilecode(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,
):
    """
    Convert a point geometry to a Tilecode grid cell.

    Args:
        resolution (int): Tilecode resolution [0..29]
        point (shapely.geometry.Point): Point geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for points)
        compact (bool, optional): Enable Tilecode compact mode (not used for points)
        topology (bool, optional): Enable topology preserving mode (not used for points)
        include_properties (bool, optional): Whether to include properties in output
        all_points (list, optional): List of points for topology preservation

    Returns:
        list: List of GeoJSON feature dictionaries representing Tilecode cells containing the point
    """
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        shortest_distance = shortest_point_distance(all_points)
        tilecode_cell_sizes = [40075016.68557849 / (2**z) for z in range(30)]
        for res in range(0, 30):
            cell_diameter = tilecode_cell_sizes[res] * sqrt(2) * 2
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 29
    tilecode_features = []
    tilecode_id = tilecode.latlon2tilecode(point.y, point.x, resolution)
    tilecode_cell = mercantile.tile(point.x, point.y, resolution)
    bounds = mercantile.bounds(tilecode_cell)
    if bounds:
        min_lat, min_lon = bounds.south, bounds.west
        max_lat, max_lon = bounds.north, bounds.east
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],
                [max_lon, min_lat],
                [max_lon, max_lat],
                [min_lon, max_lat],
                [min_lon, min_lat],
            ]
        )
        tilecode_feature = graticule_dggs_to_feature(
            "tilecode", tilecode_id, resolution, cell_polygon
        )
        if include_properties and feature_properties:
            tilecode_feature["properties"].update(feature_properties)
        tilecode_features.append(tilecode_feature)
    return tilecode_features


def polyline2tilecode(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,
):
    """
    Convert line geometries (LineString, MultiLineString) to Tilecode grid cells.

    Args:
        resolution (int): Tilecode resolution [0..29]
        feature (shapely.geometry.LineString or shapely.geometry.MultiLineString): Line geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for lines)
        compact (bool, optional): Enable Tilecode compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode (not used for lines)
        include_properties (bool, optional): Whether to include properties in output
        all_polylines (list, optional): List of polylines for topology preservation

    Returns:
        list: List of GeoJSON feature dictionaries representing Tilecode cells intersecting the line
    """
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        shortest_distance = shortest_polyline_distance(all_polylines)
        tilecode_cell_sizes = [40075016.68557849 / (2**z) for z in range(30)]
        for res in range(0, 30):
            cell_diameter = tilecode_cell_sizes[res] * sqrt(2) * 4
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 29
    tilecode_features = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)
    else:
        return []
    for polyline in polylines:
        min_lon, min_lat, max_lon, max_lat = polyline.bounds
        tilecodes = mercantile.tiles(min_lon, min_lat, max_lon, max_lat, resolution)
        tilecode_ids = []
        for tile in tilecodes:
            tilecode_id = f"z{tile.z}x{tile.x}y{tile.y}"
            tilecode_ids.append(tilecode_id)
        for tilecode_id in tilecode_ids:
            match = re.match(r"z(\d+)x(\d+)y(\d+)", tilecode_id)
            if not match:
                raise ValueError("Invalid tilecode format. Expected format: 'zXxYyZ'")
            cell_resolution = int(match.group(1))
            z = int(match.group(1))
            x = int(match.group(2))
            y = int(match.group(3))
            bounds = mercantile.bounds(x, y, z)
            if bounds:
                min_lat, min_lon = bounds.south, bounds.west
                max_lat, max_lon = bounds.north, bounds.east
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                if cell_polygon.intersects(polyline):
                    tilecode_feature = graticule_dggs_to_feature(
                        "tilecode", tilecode_id, cell_resolution, cell_polygon
                    )
                    if feature_properties:
                        tilecode_feature["properties"].update(feature_properties)
                    tilecode_features.append(tilecode_feature)
    if compact:
        tilecode_geojson = {"type": "FeatureCollection", "features": tilecode_features}
        return tilecodecompact(tilecode_geojson)["features"]
    return tilecode_features


def polygon2tilecode(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,
):
    """
    Convert polygon geometries (Polygon, MultiPolygon) to Tilecode grid cells.

    Args:
        resolution (int): Tilecode resolution [0..29]
        feature (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): Polygon geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply ('intersect', 'within', 'centroid_within', 'largest_overlap')
        compact (bool, optional): Enable Tilecode compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode (not used for polygons)
        include_properties (bool, optional): Whether to include properties in output
        all_polygons (list, optional): List of polygons for topology preservation

    Returns:
        list: List of GeoJSON feature dictionaries representing Tilecode cells based on predicate
    """
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        shortest_distance = shortest_polygon_distance(all_polygons)
        tilecode_cell_sizes = [40075016.68557849 / (2**z) for z in range(30)]
        for res in range(0, 30):
            cell_diameter = tilecode_cell_sizes[res] * sqrt(2) * 4
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 29
    tilecode_features = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)
    else:
        return []
    for polygon in polygons:
        min_lon, min_lat, max_lon, max_lat = polygon.bounds
        tilecodes = mercantile.tiles(min_lon, min_lat, max_lon, max_lat, resolution)
        tilecode_ids = []
        for tile in tilecodes:
            tilecode_id = f"z{tile.z}x{tile.x}y{tile.y}"
            tilecode_ids.append(tilecode_id)
        for tilecode_id in tilecode_ids:
            match = re.match(r"z(\d+)x(\d+)y(\d+)", tilecode_id)
            if not match:
                raise ValueError("Invalid tilecode format. Expected format: 'zXxYyZ'")
            cell_resolution = int(match.group(1))
            z = int(match.group(1))
            x = int(match.group(2))
            y = int(match.group(3))
            bounds = mercantile.bounds(x, y, z)
            if bounds:
                min_lat, min_lon = bounds.south, bounds.west
                max_lat, max_lon = bounds.north, bounds.east
                cell_polygon = Polygon(
                    [
                        [min_lon, min_lat],
                        [max_lon, min_lat],
                        [max_lon, max_lat],
                        [min_lon, max_lat],
                        [min_lon, min_lat],
                    ]
                )
                if check_predicate(cell_polygon, polygon, predicate):
                    tilecode_feature = graticule_dggs_to_feature(
                        "tilecode", tilecode_id, cell_resolution, cell_polygon
                    )
                    if include_properties and feature_properties:
                        tilecode_feature["properties"].update(feature_properties)
                    tilecode_features.append(tilecode_feature)
    if compact:
        tilecode_geojson = {"type": "FeatureCollection", "features": tilecode_features}
        return tilecodecompact(tilecode_geojson)["features"]
    return tilecode_features


def geometry2tilecode(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a list of geometries to Tilecode grid cells.

    Args:
        geometries (shapely.geometry.BaseGeometry or list): Single geometry or list of geometries
        resolution (int): Tilecode resolution [0..29]
        properties_list (list, optional): List of property dictionaries for each geometry
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Tilecode compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Tilecode grid cells
    """
    resolution = validate_tilecode_resolution(resolution)

    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    tilecode_features = []
    for i, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue
        props = properties_list[i] if i < len(properties_list) else {}
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            tilecode_features.extend(
                point2tilecode(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                tilecode_features.extend(
                    point2tilecode(
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,  # Pass all points for topology preservation
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            tilecode_features.extend(
                polyline2tilecode(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,  # Pass all polylines for topology preservation
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            tilecode_features.extend(
                polygon2tilecode(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons=all_polygons,  # Pass all polygons for topology preservation
                )
            )
    return {"type": "FeatureCollection", "features": tilecode_features}


def dataframe2tilecode(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a pandas DataFrame with geometry column to Tilecode grid cells.

    Args:
        df (pandas.DataFrame): DataFrame with geometry column
        resolution (int): Tilecode resolution [0..29]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Tilecode compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Tilecode grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2tilecode(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2tilecode(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a GeoDataFrame to Tilecode grid cells.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to convert
        resolution (int): Tilecode resolution [0..29]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable Tilecode compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with Tilecode grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2tilecode(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def vector2tilecode(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to Tilecode grid cells from various input formats.

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): Tilecode resolution [0..29]
        compact (bool): Enable Tilecode compact mode for polygons (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions
    Returns:
        dict or str: Output in the specified format
    """
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        result = geodataframe2tilecode(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        result = dataframe2tilecode(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        result = geometry2tilecode(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2tilecode(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2tilecode(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert Tilecode result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since Tilecode uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2tilecode.gpkg", driver="GPKG")
            return "vector2tilecode.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2tilecode.parquet", index=False)
            return "vector2tilecode.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2tilecode.shp", driver="ESRI Shapefile")
            return "vector2tilecode.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2tilecode_cli():
    """
    Command-line interface for vector2tilecode conversion.

    Usage:
        python vector2tilecode.py -i input.shp -r 10 -c -f geojson -o output.geojson

    Arguments:
        -i, --input: Input file path or URL
        -r, --resolution: Tilecode resolution [0..29]
        -c, --compact: Enable Tilecode compact mode
        -p, --predicate: Spatial predicate (intersect, within, centroid_within, largest_overlap)
        -t, --topology: Enable topology preserving mode
        -np, --no-props: Do not include original feature properties
        -f, --format: Output format (geojson, gpkg, parquet, csv, shapefile)
        -o, --output: Output file path
    """
    parser = argparse.ArgumentParser(
        description="Convert vector data to Tilecode grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(0, 30),
        metavar="[0-29]",
        help="Tilecode resolution [0..29] (0=coarsest, 29=finest)",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable Tilecode compact mode for polygons",
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_tilecode_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)
    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2tilecode{extensions.get(args.format, '')}"
    try:
        vector2tilecode(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2tilecode_cli()



#!/usr/bin/env python3
"""
GeoJSON to CSV Converter

This script converts GeoJSON files to CSV format, extracting coordinates and properties
from GeoJSON features into a tabular format.
"""

import csv
import sys
import argparse
import os
from pathlib import Path
from typing import Dict, Any
from io import StringIO
from urllib.parse import urlparse
from vgrid.utils.download import is_url, read_geojson_file


def extract_coordinates(geometry: Dict[str, Any]) -> Dict[str, float]:
    """
    Extract coordinates from a GeoJSON geometry object.

    Args:
        geometry (Dict[str, Any]): GeoJSON geometry object

    Returns:
        Dict[str, float]: Dictionary containing x and y coordinates
    """
    if geometry["type"] == "Point":
        return {"x": geometry["coordinates"][0], "y": geometry["coordinates"][1]}
    elif geometry["type"] in ["LineString", "MultiPoint"]:
        # For LineString and MultiPoint, use the first point
        return {"x": geometry["coordinates"][0][0], "y": geometry["coordinates"][0][1]}
    elif geometry["type"] in ["Polygon", "MultiLineString"]:
        # For Polygon and MultiLineString, use the first point of the first ring/line
        return {
            "x": geometry["coordinates"][0][0][0],
            "y": geometry["coordinates"][0][0][1],
        }
    elif geometry["type"] == "MultiPolygon":
        # For MultiPolygon, use the first point of the first ring of the first polygon
        return {
            "x": geometry["coordinates"][0][0][0][0],
            "y": geometry["coordinates"][0][0][0][1],
        }
    else:
        raise ValueError(f"Unsupported geometry type: {geometry['type']}")


def geojson2csv(geojson_data):
    """
    Convert GeoJSON data to CSV format.

    Args:
        geojson_data (Dict): GeoJSON data as a dictionary with 'type' and 'features' keys

    Returns:
        str: CSV data as a string

    Raises:
        ValueError: If the GeoJSON format is invalid or contains no features
    """
    features = geojson_data["features"]
    if not features:
        raise ValueError("No features found in GeoJSON data")

    # Prepare CSV headers
    headers = ["x", "y"]  # Always include coordinates
    if features[0].get("properties"):
        headers.extend(features[0]["properties"].keys())

    # Create CSV in memory
    output = StringIO()
    writer = csv.DictWriter(output, fieldnames=headers)
    writer.writeheader()

    for feature in features:
        row = extract_coordinates(feature["geometry"])
        if feature.get("properties"):
            row.update(feature["properties"])
        writer.writerow(row)

    return output.getvalue()


def geojson2csv_cli() -> None:
    """
    Command-line interface for converting GeoJSON to CSV.

    Usage:
        python geojson2csv.py -geojson input.geojson
        python geojson2csv.py -geojson https://example.com/data.geojson
    """
    parser = argparse.ArgumentParser(description="Convert GeoJSON to CSV format")
    parser.add_argument(
        "-geojson", required=True, help="Input GeoJSON file path or URL"
    )

    args = parser.parse_args()

    # Read GeoJSON data from file or URL
    geojson_data = read_geojson_file(args.geojson)
    if geojson_data is None:
        sys.exit(1)

    try:
        csv_data = geojson2csv(geojson_data)

        # Create output CSV path
        if is_url(args.geojson):
            # For URLs, use the last part of the URL as the base filename
            base_name = os.path.splitext(os.path.basename(urlparse(args.geojson).path))[
                0
            ]
            if not base_name:  # If URL doesn't end with a filename
                base_name = "geojson_data"
            output_path = f"{base_name}.csv"
        else:
            # For local files, replace .geojson extension with .csv
            input_path = Path(args.geojson)
            output_path = input_path.with_suffix(".csv")

        # Write CSV data to file
        with open(output_path, "w", newline="", encoding="utf-8") as f:
            f.write(csv_data)
        print(f"Successfully converted to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}")
        sys.exit(1)


if __name__ == "__main__":
    geojson2csv_cli()



import sys
import argparse
import json
from tqdm import tqdm
import pandas as pd
import geopandas as gpd
from shapely.geometry import MultiPoint, MultiLineString, MultiPolygon
from vgrid.conversion.latlon2dggs import latlon2mgrs
from vgrid.conversion.dggs2geojson import mgrs2geojson
from vgrid.utils.geometry import (
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)
from math import sqrt


def validate_mgrs_resolution(resolution):
    """
    Validate that MGRS resolution is in the valid range [0..5] (0=coarsest, 5=finest).

    Args:
        resolution: Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [0..5]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution < 0 or resolution > 5:
        raise ValueError(f"Resolution must be in range [0..5], got {resolution}")

    return resolution


def point2mgrs(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,
):
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        shortest_distance = shortest_point_distance(all_points)
        mgrs_cell_sizes = [100000, 10000, 1000, 100, 10, 1]
        for res in range(0, 6):
            cell_diameter = mgrs_cell_sizes[res] * sqrt(2) * 2
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 5
    mgrs_features = []
    latitude, longitude = point.y, point.x
    mgrs_id = latlon2mgrs(latitude, longitude, resolution)
    mgrs_feature = mgrs2geojson(mgrs_id)
    if mgrs_feature:
        if include_properties and feature_properties:
            mgrs_feature["properties"].update(feature_properties)
        mgrs_features.append(mgrs_feature)
    return mgrs_features


def polyline2mgrs(
    resolution,
    geometry,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,
):
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        shortest_distance = shortest_polyline_distance(all_polylines)
        mgrs_cell_sizes = [100000, 10000, 1000, 100, 10, 1]
        for res in range(0, 6):
            cell_diameter = mgrs_cell_sizes[res] * sqrt(2) * 4
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 5
    return []


def polygon2mgrs(
    resolution,
    geometry,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,
):
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        shortest_distance = shortest_polygon_distance(all_polygons)
        mgrs_cell_sizes = [100000, 10000, 1000, 100, 10, 1]
        for res in range(0, 6):
            cell_diameter = mgrs_cell_sizes[res] * sqrt(2) * 4
            if cell_diameter < shortest_distance:
                resolution = res
                break
        else:
            resolution = 5
    return []


def geometry2mgrs(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    resolution = validate_mgrs_resolution(resolution)

    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    mgrs_features = []
    for i, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue
        props = properties_list[i] if i < len(properties_list) else {}
        if not include_properties:
            props = {}
        if geom.geom_type == "Point":
            mgrs_features.extend(
                point2mgrs(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
            )
        elif geom.geom_type == "MultiPoint":
            for pt in geom.geoms:
                mgrs_features.extend(
                    point2mgrs(
                        resolution,
                        pt,
                        props,
                        predicate,
                        compact,
                        topology,
                        include_properties,
                        all_points,  # Pass all points for topology preservation
                    )
                )
        elif geom.geom_type in ("LineString", "MultiLineString"):
            mgrs_features.extend(
                polyline2mgrs(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polylines,  # Pass all polylines for topology preservation
                )
            )
        elif geom.geom_type in ("Polygon", "MultiPolygon"):
            mgrs_features.extend(
                polygon2mgrs(
                    resolution,
                    geom,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_polygons=all_polygons,  # Pass all polygons for topology preservation
                )
            )

    return {"type": "FeatureCollection", "features": mgrs_features}


def dataframe2mgrs(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2mgrs(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2mgrs(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2mgrs(
        geometries, resolution, properties_list, compact, include_properties
    )


def vector2mgrs(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        result = geodataframe2mgrs(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        result = dataframe2mgrs(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        result = geometry2mgrs(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2mgrs(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2mgrs(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert MGRS result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since MGRS uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2mgrs.gpkg", driver="GPKG")
            return "vector2mgrs.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2mgrs.parquet", index=False)
            return "vector2mgrs.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2mgrs.shp", driver="ESRI Shapefile")
            return "vector2mgrs.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2mgrs_cli():
    parser = argparse.ArgumentParser(description="Convert vector data to MGRS cells")
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=range(0, 6),
        metavar="[0-5]",
        help="MGRS resolution [0..5] (0=coarsest, 5=finest)",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", help="Enable topology preserving mode"
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable MGRS compact mode for polygons",
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_mgrs_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)

    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2mgrs{extensions.get(args.format, '')}"
    try:
        vector2mgrs(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2mgrs_cli()



# Reference: https://github.com/brdo/icosahedron
"""
Example usage:
python icosahedron.py --json --dual --level 3 > icosahedron.json && ogr2ogr -f PostgreSQL PG:"user=root password=secret" -overwrite -nln icosahedron icosahedron.json
"""

import json
import math
import collections
import numpy as np

ORIGIN = np.array([0, 0, 0], dtype=np.float64)

# golden ratio
PHI = (1 + math.sqrt(5)) / 2


def distance(v1, v2=ORIGIN):
    """
    return distance between two vectors
    """
    return np.linalg.norm(v1 - v2)


def midpoint(v1, v2):
    """
    return midpoint between two vectors
    """
    v = v1 + v2
    mp = v / np.linalg.norm(v)
    return mp


class Vertice(object):
    vertices = {}

    def __init__(self, vertice, face=None):
        self.vertice = vertice
        self.key = tuple(self.vertice)
        self._faces = set()

    @property
    def face(self):
        return self._faces

    @face.setter
    def face(self, face):
        if face:
            self._faces.add(face.key)

    @classmethod
    def make(cls, vertice, face=None):
        vert = Vertice(vertice, face)
        if vert.key in Vertice.vertices:
            vert = cls.vertices[vert.key]
            return vert
        else:
            cls.vertices[vert.key] = vert
            vert.id = len(cls.vertices)
            return vert

    def distance(self, other):
        return distance(self.vertice, other.vertice)

    def latlon(self):
        """
        return lon, lat of point
        """
        x, y, z = tuple(self.vertice.tolist())
        th = 25 * math.pi / 180
        x1 = math.cos(th) * x - math.sin(th) * y
        y1 = math.sin(th) * x + math.cos(th) * y
        return [math.atan2(y1, x1) * 180 / math.pi, math.asin(z) * 180 / math.pi]

    @property
    def coordinates(self):
        return self.latlon()

    @classmethod
    def clear(cls):
        cls.vertices = {}

    def __repr__(self):
        return "Vertice(%d)" % (self.id)


class Face(object):
    faces = {}

    def __init__(self, vertices, id):
        self.id = id
        self.vertices = [Vertice.make(vert, self) for vert in vertices]
        self.key = tuple(sorted([vert.key for vert in self.vertices]))

        for vertice in self.vertices:
            vertice.face = self

        # convert face to a vertex of the polyhedral dual

        # geometrically point x lies on face with closest approach to origin y
        # IFF
        #   1. x . y = y . y
        # use polar reciprocal transformation to convert vertex x into plane x'
        #   2. x' = x / (x . x)
        #
        # in order for the transformation to be correct
        # vertex y' must lie on plane x'
        # IFF
        #   3. x' . y' = x' . x'
        # plugging Eq 2 into 3 for both x' and y' and simplifying
        #   x . y / (( y . y) ( x . x )) = x . x / (x . x) ^2
        #   x . y / (y . y) = 1
        #   x . y = y . y
        #
        # which proves that vertex x lies on plane y under the tranformation
        # so we've successfully shown that the polar reciprocal tranformation
        # can be used to tranform a polyhedral into its dual

        # normal can be computed using (v2 - v1) cross (v3 - v1)
        normal = np.cross(self.v2 - self.v1, self.v3 - self.v1)
        self.normal = normal / np.linalg.norm(normal)

        # plugging Eq 1. into Eq 2.
        # x' = x / (x . y)
        self.vertex = self.normal / (
            sum([np.dot(v, self.normal) for v in vertices]) / len(vertices)
        )
        self.vertex = self.vertex / np.linalg.norm(self.vertex)

    @classmethod
    def make(cls, vertices):
        id = len(cls.faces) + 1
        face = Face(vertices, id)
        if face.key in cls.faces:
            return cls.faces[face.key]
        else:
            cls.faces[face.key] = face
            return face

    @classmethod
    def clear(cls):
        cls.faces = {}

    @property
    def v1(self):
        return self.vertices[0].vertice

    @property
    def v2(self):
        return self.vertices[1].vertice

    @property
    def v3(self):
        return self.vertices[2].vertice

    @property
    def coordinates(self):
        coordinates = [vert.coordinates for vert in self.vertices]

        coordinates.append(coordinates[0])

        l = [c[0] for c in coordinates]
        dl = max(l) - min(l)
        if dl > 180:
            coordinates = [[c[0] + 360, c[1]] if c[0] < 0 else c for c in coordinates]

        return coordinates

    def __repr__(self):
        return "Face(%d)" % (self.id)


def icosahedron_vertices():
    """
    generate icosahedron vertices
    """
    # used to permute for all 12 vertices
    point = np.array([0, 1, PHI], dtype=np.float64)

    # normalize
    point = point / distance(point)

    # construct 12 vertices by cyclically permuting [0, +/-1, +/- PHI]
    pqueue = collections.deque(point)

    Vertice.clear()

    for _ in range(3):
        pqueue.rotate()
        for i in [-1, 1]:
            for j in [-1, 1]:
                for k in [-1, 1]:
                    Vertice.make(np.multiply(pqueue, [i, j, k]))


def icosahedron_faces():
    """
    generate icosahedron faces
    """
    pairs = collections.defaultdict(set)
    for vert in Vertice.vertices.values():
        min_dist = min(
            [
                vert.distance(other_vert)
                for other_vert in Vertice.vertices.values()
                if vert.distance(other_vert) > 0
            ]
        )

        for other_vert in Vertice.vertices.values():
            if vert.distance(other_vert) == min_dist:
                pairs[vert.key].add(other_vert.key)
                pairs[other_vert.key].add(vert.key)

    Face.clear()

    # Change items() to items() for Python 3 compatibility
    for vert, verts in pairs.items():
        for v1 in verts:
            if v1 == vert:
                continue
            for v2 in verts:
                if v2 == vert or v2 == v1:
                    continue
                if (
                    v2 in pairs[v1]
                    and v1 in pairs[v2]
                    and vert in pairs[v1]
                    and vert in pairs[v2]
                ):
                    key = tuple(sorted([vert, v1, v2]))

                    face = Face.make([np.array(v, dtype=np.float64) for v in key])

                    Vertice.vertices.get(vert).face = face
                    Vertice.vertices.get(v1).face = face
                    Vertice.vertices.get(v2).face = face


def icosahedron_bisect(level):
    """
    bisect the faces "level" times
    """

    l = 1
    while l < level:
        l += 1

        faces = Face.faces

        Vertice.clear()
        Face.clear()

        for fkey, face in faces.items():
            a = midpoint(face.v1, face.v2)
            b = midpoint(face.v2, face.v3)
            c = midpoint(face.v3, face.v1)

            Face.make([a, face.v1, c])
            Face.make([a, face.v2, b])
            Face.make([a, b, c])
            Face.make([b, face.v3, c])


def icosahedron_geojson():
    """
    return geojson
    """
    json = {"type": "FeatureCollection", "features": []}

    for fkey, face in Face.faces.items():
        vert = face.vertex
        json["features"].append(
            dict(
                type="Feature",
                geometry=dict(type="Polygon", coordinates=[face.coordinates]),
                properties=dict(x=vert[0], y=vert[1], z=vert[2]),
            )
        )

    return json


def icosahedron_dual():
    """
    convert icosahedron into its dual
    """

    # build a mapping of vertices to the connecting faces
    vertFaces = collections.defaultdict(set)
    for vert in Vertice.vertices.values():
        for fkey in vert.face:
            vertFaces[vert.key].add(fkey)

    # store the vertices and faces
    faces = Face.faces

    # clear the vertices and faces for the dual
    Vertice.clear()
    Face.clear()

    # build the new vertices and faces
    for vkey, fkeys in vertFaces.items():
        # store the first face
        face = faces[fkeys.pop()]
        _faces = [face]
        l = len(fkeys)

        # loop through the other faces
        # grabbing the nearest face (by its corresponding vertex)
        while len(_faces) <= l:
            fkey = sorted(
                fkeys, key=lambda fkey: distance(face.vertex, faces[fkey].vertex)
            )[0]
            face = faces[fkey]
            fkeys.remove(fkey)
            _faces.append(faces[fkey])

        # build a new face from the plane -> vertices of the original faces
        Face.make([face.vertex for face in _faces])


def main():
    """
    run the program
    """
    from optparse import OptionParser

    parser = OptionParser()

    parser.add_option(
        "--dual",
        dest="dual",
        action="store_true",
        default=True,
        help="Convert polyhedron to its dual",
    )

    parser.add_option(
        "--json", dest="json", action="store_true", help="Display GeoJSON to stdout"
    )

    parser.add_option(
        "--level", dest="level", type="int", default=1.0, help="Bifurcation level"
    )

    options, args = parser.parse_args()

    icosahedron_vertices()

    icosahedron_faces()

    icosahedron_bisect(options.level)

    icosahedron_dual()

    if options.json:
        print(json.dumps(icosahedron_geojson()))


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint


## Unit tests for the LatLongPoint class.
class TestLatLongPoint(unittest.TestCase):
    def test_getters(self):
        latitude = 1.0
        longitude = 2.0
        accuracy = 3.0
        point = LatLongPoint(latitude, longitude, accuracy)
        self.assertEqual(point.get_latitude(), latitude)
        self.assertEqual(point.get_longitude(), longitude)
        self.assertEqual(point.get_accuracy(), accuracy)

    def test_operators(self):
        orig_point = LatLongPoint(1.0, 2.0, 3.0)
        same_point = LatLongPoint(1.0, 2.0, 3.0)
        diff_point = LatLongPoint(4.0, 5.0, 6.0)
        self.assertTrue(orig_point == same_point)
        self.assertFalse(orig_point == diff_point)
        self.assertEqual(orig_point, same_point)
        self.assertFalse(orig_point != same_point)
        self.assertTrue(orig_point != diff_point)
        self.assertNotEqual(orig_point, diff_point)



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""


## Checks the input corresponds to one of the DggsShapeLocation values.
#  @param value Input to check.
#  @throw ValueError Thrown if the input is not a DGGS shape location.
def check_dggs_shape_location(value):
    if not (
        isinstance(value, int)
        and value >= DggsShapeLocation.ONE_FACE
        and value <= DggsShapeLocation.NO_LOCATION
    ):
        raise ValueError(
            "DGGS shape location must be one of the DggsShapeLocation values"
        )


## Different DGGS shape location types.
class DggsShapeLocation:
    ## Shape sits entirely on one face.
    ONE_FACE = 0
    ## Shape spans two faces.
    TWO_FACES = 1
    ## Shape spans more than two faces.
    MANY_FACES = 2
    ## Shape location only applies to cells - this value is used for linestrings and polygons.
    NO_LOCATION = 3



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.shapes.lat_long_point import LatLongPoint
from vgrid.utils.eaggr.shapes.lat_long_linestring import LatLongLinestring
from vgrid.utils.eaggr.shapes.lat_long_polygon import LatLongPolygon


## Unit tests for the LatLongPolygon class.
class LatLongPolygonTest(unittest.TestCase):
    def test_getters(self):
        # Create LatLongPolygon object
        orig_points = [
            LatLongPoint(1.0, 2.0, 3.0),
            LatLongPoint(4.0, 5.0, 6.0),
            LatLongPoint(7.0, 8.0, 9.0),
        ]
        linestring = LatLongLinestring()
        linestring.add_point(orig_points[0])
        polygon = LatLongPolygon(linestring)
        linestring = LatLongLinestring()
        linestring.add_point(orig_points[1])
        polygon.add_inner_ring(linestring)
        linestring = LatLongLinestring()
        linestring.add_point(orig_points[2])
        polygon.add_inner_ring(linestring)
        # Test LatLongPolygon object
        outer_ring_points = polygon.get_outer_ring().get_points()
        self.assertEqual(outer_ring_points[0], orig_points[0])
        inner_rings = polygon.get_inner_rings()
        self.assertEqual(len(inner_rings), 2)
        inner_ring_points = inner_rings[0].get_points()
        self.assertEqual(inner_ring_points[0], orig_points[1])
        inner_ring_points = inner_rings[1].get_points()
        self.assertEqual(inner_ring_points[0], orig_points[2])

    def test_errors(self):
        # Error with outer ring
        with self.assertRaises(ValueError):
            LatLongPolygon("Not a LatLongLinestring")
        # Error with inner ring
        polygon = LatLongPolygon(LatLongLinestring())
        with self.assertRaises(ValueError):
            polygon.add_inner_ring("Not a LatLongLinestring")



import os
import argparse
import json
from tqdm import tqdm
import rasterio
import numpy as np
import csv
from vgrid.stats.rhealpixstats import rhealpix_metrics
from vgrid.generator.settings import geodesic_dggs_metrics, geodesic_dggs_to_feature
from math import cos, radians
from vgrid.utils.rhealpixdggs.dggs import RHEALPixDGGS
from vgrid.utils.rhealpixdggs.ellipsoids import WGS84_ELLIPSOID
from vgrid.conversion.dggs2geojson import rhealpix_cell_to_polygon


def get_nearest_rhealpix_resolution(raster_path):
    with rasterio.open(raster_path) as src:
        transform = src.transform
        crs = src.crs
        pixel_width = transform.a
        pixel_height = -transform.e
        cell_size = pixel_width * pixel_height

        if crs.is_geographic:
            # Latitude of the raster center
            center_latitude = (src.bounds.top + src.bounds.bottom) / 2
            # Convert degrees to meters
            meter_per_degree_lat = 111_320  # Roughly 1 degree latitude in meters
            meter_per_degree_lon = meter_per_degree_lat * cos(radians(center_latitude))

            pixel_width_m = pixel_width * meter_per_degree_lon
            pixel_height_m = pixel_height * meter_per_degree_lat
            cell_size = pixel_width_m * pixel_height_m

    nearest_resolution = None
    min_diff = float("inf")

    # Check resolutions from 0 to 15
    for res in range(16):
        _, _, avg_area = rhealpix_metrics(res)
        diff = abs(avg_area - cell_size)
        # If the difference is smaller than the current minimum, update the nearest resolution
        if diff < min_diff:
            min_diff = diff
            nearest_resolution = res

    return nearest_resolution


def convert_numpy_types(obj):
    """Recursively convert NumPy types to native Python types"""
    if isinstance(obj, np.generic):
        return obj.item()  # Convert numpy types like np.uint8 to native Python int
    elif isinstance(obj, dict):
        return {k: convert_numpy_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_numpy_types(v) for v in obj]
    else:
        return obj


def raster2rhealpix(rhealpix_dggs, raster_path, resolution=None, format="geojson"):
    # Step 1: Determine the nearest rhealpix resolution if none is provided
    if resolution is None:
        resolution = get_nearest_rhealpix_resolution(raster_path)
        print(f"Nearest rhealpix resolution determined: {resolution}")

    # Open the raster file to get metadata and data
    with rasterio.open(raster_path) as src:
        raster_data = src.read()  # Read all bands
        transform = src.transform
        width, height = src.width, src.height
        band_count = src.count  # Number of bands in the raster

    rhealpix_ids = set()

    for row in range(height):
        for col in range(width):
            lon, lat = transform * (col, row)
            point = (lon, lat)
            rhealpix_cell = rhealpix_dggs.cell_from_point(
                resolution, point, plane=False
            )
            rhealpix_ids.add(str(rhealpix_cell))

    # Sample the raster values at the centroids of the rhealpix hexagons
    rhealpix_data = []

    for rhealpix_id in tqdm(rhealpix_ids, desc="Resampling", unit=" cells"):
        rhealpix_uids = (rhealpix_id[0],) + tuple(map(int, rhealpix_id[1:]))
        rhealpix_cell = rhealpix_dggs.cell(rhealpix_uids)
        cell_polygon = rhealpix_cell_to_polygon(rhealpix_cell)
        num_edges = 4
        if rhealpix_cell.ellipsoidal_shape() == "dart":
            num_edges = 3
        centroid_lat, centroid_lon, avg_edge_len, cell_area = geodesic_dggs_metrics(
            cell_polygon, num_edges
        )

        col, row = ~transform * (centroid_lon, centroid_lat)

        if 0 <= col < width and 0 <= row < height:
            # Get the values for all bands at this centroid
            values = raster_data[:, int(row), int(col)]
            rhealpix_data.append(
                {
                    "rhealpix": rhealpix_id,
                    **{
                        f"band_{i + 1}": values[i] for i in range(band_count)
                    },  # Create separate columns for each band
                }
            )

    if format.lower() == "csv":
        import io

        output = io.StringIO()
        if rhealpix_data:
            writer = csv.DictWriter(output, fieldnames=rhealpix_data[0].keys())
            writer.writeheader()
            writer.writerows(rhealpix_data)
        return output.getvalue()

    # Create the GeoJSON-like structure
    rhealpix_features = []
    for data in tqdm(rhealpix_data, desc="Converting to GeoJSON", unit=" cells"):
        rhealpix_uids = (data["rhealpix"][0],) + tuple(map(int, data["rhealpix"][1:]))
        rhealpix_cell = rhealpix_dggs.cell(rhealpix_uids)
        if rhealpix_cell:
            cell_polygon = rhealpix_cell_to_polygon(rhealpix_cell)
            num_edges = 4
            if rhealpix_cell.ellipsoidal_shape() == "dart":
                num_edges = 3
            rhealpix_feature = geodesic_dggs_to_feature(
                "rhealpix", str(rhealpix_cell), resolution, cell_polygon, num_edges
            )
            band_properties = {
                f"band_{i + 1}": data[f"band_{i + 1}"] for i in range(band_count)
            }
            rhealpix_feature["properties"].update(convert_numpy_types(band_properties))
            rhealpix_features.append(rhealpix_feature)

    return {"type": "FeatureCollection", "features": rhealpix_features}


def raster2rhealpix_cli():
    """Command line interface for raster2rhealpix"""
    parser = argparse.ArgumentParser(
        description="Convert Raster in Geographic CRS to rHEALPix DGGS"
    )
    parser.add_argument("-raster", type=str, required=True, help="Raster file path")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        required=False,
        default=None,
        help="Resolution [0..15]",
    )
    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=False,
        default="geojson",
        choices=["geojson", "csv"],
        help="Output format (geojson or csv)",
    )

    args = parser.parse_args()

    E = WGS84_ELLIPSOID
    rhealpix_dggs = RHEALPixDGGS(ellipsoid=E, north_square=1, south_square=3, N_side=3)

    if not os.path.exists(args.raster):
        raise FileNotFoundError(f"The file {args.raster} does not exist.")

    if args.resolution is not None and (args.resolution < 0 or args.resolution > 15):
        raise ValueError("Resolution must be in range [0..15]")

    result = raster2rhealpix(rhealpix_dggs, args.raster, args.resolution, args.format)

    # Generate output filename
    base_name = os.path.splitext(os.path.basename(args.raster))[0]

    if args.format.lower() == "csv":
        output_path = f"{base_name}2rhealpix.csv"
        with open(output_path, "w", newline="") as f:
            f.write(result)
    else:
        output_path = f"{base_name}2rhealpix.geojson"
        with open(output_path, "w") as f:
            json.dump(result, f)

    print(f"Output saved as {output_path}")



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""


## Exception thrown if a DGGS DLL function fails.
class EaggrException(Exception):
    ## Stores an error message and the value returned by the DGGS DLL function.
    def __init__(self, return_code, message):
        # Call the base class constructor with the parameters it needs
        super(EaggrException, self).__init__(message)
        # Store the return code given by the DGGS DLL function
        self._return_code = return_code

    ## @return Return code given by the DGGS API function.
    def get_return_code(self):
        return self._return_code



import shapely
from pyproj import Geod
from shapely.geometry import Polygon, Point, LineString
import numpy as np
import matplotlib.pyplot as plt
import geopandas as gpd

# Initialize Geod with WGS84 ellipsoid
geod = Geod(ellps="WGS84")

def shortest_point_distance(points):
    """
    Calculate distances between points in a Shapely geometry.
    If there's only one point, return 0.
    If there are multiple points, calculate Delaunay triangulation and return distances.

    Args:
        points: Shapely Point or MultiPoint geometry

    Returns:
        tuple: shortest_distance
    """
    # Handle single Point
    if isinstance(points, Point):
        return 0  # Single point has no distance to other points

    # Handle MultiPoint with single point
    if len(points.geoms) == 1:
        return 0

    # Generate Delaunay triangulation
    delaunay = shapely.delaunay_triangles(points, only_edges=True)

    # Find the shortest edge
    shortest_distance = float("inf")

    for line in delaunay.geoms:
        # Get the coordinates of the line endpoints
        coords = list(line.coords)
        lon1, lat1 = coords[0]
        lon2, lat2 = coords[1]

        # Calculate the distance in meters using pyproj Geod
        distance = geod.inv(lon1, lat1, lon2, lat2)[2]  # [2] gives the distance in meters
        if distance < shortest_distance:
            shortest_distance = distance

    return shortest_distance if shortest_distance != float("inf") else 0


def shortest_polyline_distance(polylines):
    """
    Calculate the shortest distance between polylines using GeoPandas shortest_line() method.
    If there's only one polyline, return 0.
    If there are multiple polylines, use shortest_line() and return the shortest distance in meters.

    Args:
        polylines: Shapely LineString or MultiLineString geometry, or GeoSeries of LineStrings
        
    Returns:
        float: shortest_distance between polylines in meters
    """
    # Handle single LineString
    if isinstance(polylines, LineString):
        return 0  # Single polyline has no distance to other polylines
    
    # Handle MultiLineString with single line
    if hasattr(polylines, 'geoms') and len(polylines.geoms) == 1:
        return 0
    
    # Handle GeoSeries
    if hasattr(polylines, 'iloc'):
        # Already a GeoSeries
        line_list = list(polylines.geometry)
        gs = polylines
    else:
        # Handle MultiLineString or list
        line_list = list(polylines.geoms) if hasattr(polylines, 'geoms') else [polylines]
        gs = gpd.GeoSeries(line_list)
    
    if len(line_list) < 2:
        return 0
    
    # Calculate shortest distance between all pairs of polylines using shortest_line()
    shortest_distance = float("inf")
    
    for i in range(len(line_list)):
        for j in range(i+1, len(line_list)):
            line1 = line_list[i]
            line2 = line_list[j]
            
            # Check if polylines are disjoint
            if line1.disjoint(line2):
                try:
                    # Create GeoSeries for shortest_line calculation
                    gs1 = gpd.GeoSeries([line1])
                    gs2 = gpd.GeoSeries([line2])
                    
                    # Get shortest line using GeoPandas method
                    shortest_line = gs1.shortest_line(gs2, align=False).iloc[0]
                    
                    if shortest_line and shortest_line.length > 0:
                            # Get the endpoints of the shortest line
                            coords = list(shortest_line.coords)
                            if len(coords) >= 2:
                                lon1, lat1 = coords[0]
                                lon2, lat2 = coords[1]
                                
                                # Calculate geodesic distance in meters
                                distance = geod.inv(lon1, lat1, lon2, lat2)[2]  # [2] gives distance in meters
                                
                                if distance < shortest_distance:
                                    shortest_distance = distance
                except Exception as e:
                    print(f"Error calculating distance between polylines {i} and {j}: {e}")
                    continue
    
    return shortest_distance if shortest_distance != float("inf") else 0 


def shortest_polygon_distance(polygons):
    """
    Calculate the shortest distance between polygons using GeoPandas shortest_line() method.
    If there's only one polygon, return 0.
    If there are multiple polygons, use shortest_line() and return the shortest distance in meters.

    Args:
        polygons: Shapely Polygon or MultiPolygon geometry, or GeoSeries of Polygons
        
    Returns:
        float: shortest_distance between polygons in meters
    """
    # Handle single Polygon
    if isinstance(polygons, Polygon):
        return 0  # Single polygon has no distance to other polygons
    
    # Handle MultiPolygon with single polygon
    if hasattr(polygons, 'geoms') and len(polygons.geoms) == 1:
        return 0
    
    # Handle GeoSeries
    if hasattr(polygons, 'iloc'):
        # Already a GeoSeries
        polygon_list = list(polygons.geometry)
        gs = polygons
    else:
        # Handle MultiPolygon or list
        polygon_list = list(polygons.geoms) if hasattr(polygons, 'geoms') else [polygons]
        gs = gpd.GeoSeries(polygon_list)
    
    if len(polygon_list) < 2:
        return 0
    
    # Calculate shortest distance between all pairs of polygons using shortest_line()
    shortest_distance = float("inf")
    
    for i in range(len(polygon_list)):
        for j in range(i+1, len(polygon_list)):
            polygon1 = polygon_list[i]
            polygon2 = polygon_list[j]
            
            # Check if polygons are disjoint
            if polygon1.disjoint(polygon2):
                try:
                    # Create GeoSeries for shortest_line calculation
                    gs1 = gpd.GeoSeries([polygon1])
                    gs2 = gpd.GeoSeries([polygon2])
                    
                    # Get shortest line using GeoPandas method
                    shortest_line = gs1.shortest_line(gs2, align=False).iloc[0]
                    
                    if shortest_line and shortest_line.length > 0:
                            # Get the endpoints of the shortest line
                            coords = list(shortest_line.coords)
                            if len(coords) >= 2:
                                lon1, lat1 = coords[0]
                                lon2, lat2 = coords[1]
                                
                                # Calculate geodesic distance in meters
                                distance = geod.inv(lon1, lat1, lon2, lat2)[2]  # [2] gives distance in meters
                                
                                if distance < shortest_distance:
                                    shortest_distance = distance
                except Exception as e:
                    print(f"Error calculating distance between polygons {i} and {j}: {e}")
                    continue
    
    return shortest_distance if shortest_distance != float("inf") else 0


def geodesic_distance(
    lat: float, lon: float, length_meter: float
) -> tuple[float, float]:
    """
    Convert meters to approximate degree offsets at a given location.

    Parameters:
        lat (float): Latitude of the reference point
        lon (float): Longitude of the reference point
        length_meter (float): Distance in meters

    Returns:
        (delta_lat_deg, delta_lon_deg): Tuple of degree offsets in latitude and longitude
    """
    # Move north for latitude delta
    lon_north, lat_north, _ = geod.fwd(lon, lat, 0, length_meter)
    delta_lat = lat_north - lat

    # Move east for longitude delta
    lon_east, lat_east, _ = geod.fwd(lon, lat, 90, length_meter)
    delta_lon = lon_east - lon

    return delta_lat, delta_lon


def geodesic_buffer(polygon, distance):
    """
    Create a geodesic buffer around a polygon using pyproj Geod.

    Args:
        polygon: Shapely Polygon geometry
        distance: Buffer distance in meters

    Returns:
        Shapely Polygon: Buffered polygon
    """
    buffered_coords = []
    for lon, lat in polygon.exterior.coords:
        # Generate points around the current vertex to approximate a circle
        circle_coords = [
            geod.fwd(lon, lat, azimuth, distance)[
                :2
            ]  # Forward calculation: returns (lon, lat, back_azimuth)
            for azimuth in range(0, 360, 10)  # Generate points every 10 degrees
        ]
        buffered_coords.append(circle_coords)

    # Flatten the list of buffered points and form a Polygon
    all_coords = [coord for circle in buffered_coords for coord in circle]
    return Polygon(all_coords).convex_hull


def check_predicate(cell_polygon, input_geometry, predicate=None):
    """
    Determine whether to keep an H3 cell based on its relationship with the input geometry.

    Args:
        cell_polygon: Shapely Polygon representing the H3 cell
        input_geometry: Shapely geometry (Polygon, LineString, etc.)
        predicate (str or int): Spatial predicate to apply:
            String values:
                None or "intersects": intersects (default)
                "within": within
                "centroid_within": centroid_within
                "largest_overlap": intersection >= 50% of cell area
            Integer values (for backward compatibility):
                None or 0: intersects (default)
                1: within
                2: centroid_within
                3: intersection >= 50% of cell area

    Returns:
        bool: True if cell should be kept, False otherwise
    """
    # Handle string predicates
    if isinstance(predicate, str):
        predicate_lower = predicate.lower()
        if predicate_lower in ["intersects", "intersect"]:
            return cell_polygon.intersects(input_geometry)
        elif predicate_lower == "within":
            return cell_polygon.within(input_geometry)
        elif predicate_lower in ["centroid_within", "centroid"]:
            return cell_polygon.centroid.within(input_geometry)
        elif predicate_lower in ["largest_overlap", "overlap", "majority"]:
            # intersection >= 50% of cell area
            if cell_polygon.intersects(input_geometry):
                intersection_geom = cell_polygon.intersection(input_geometry)
                if intersection_geom and intersection_geom.area > 0:
                    intersection_area = intersection_geom.area
                    cell_area = cell_polygon.area
                    return (intersection_area / cell_area) >= 0.5
            return False
        else:
            # Unknown string predicate, default to intersects
            return cell_polygon.intersects(input_geometry)

    # Handle integer predicates (backward compatibility)
    elif isinstance(predicate, int):
        if predicate == 0:
            # Default: intersects
            return cell_polygon.intersects(input_geometry)
        elif predicate == 1:
            # within
            return cell_polygon.within(input_geometry)
        elif predicate == 2:
            # centroid_within
            return cell_polygon.centroid.within(input_geometry)
        elif predicate == 3:
            # intersection >= 50% of cell area
            if cell_polygon.intersects(input_geometry):
                intersection_geom = cell_polygon.intersection(input_geometry)
                if intersection_geom and intersection_geom.area > 0:
                    intersection_area = intersection_geom.area
                    cell_area = cell_polygon.area
                    return (intersection_area / cell_area) >= 0.5
            return False
        else:
            # Unknown predicate, default to intersects
            return cell_polygon.intersects(input_geometry)

    else:
        # None or other types, default to intersects
        return cell_polygon.intersects(input_geometry)


#### Test - Shortest Lines Between Polygons
# Example demonstrating shortest lines between polygons

# Create test polygons (3 separated polygons)
# polygons = [
#     Polygon([(0, 0), (2, 2), (0, 2)]),
#     Polygon([(4, 4), (7, 8), (8, 0)]),
#     Polygon([(8, 8), (10, 12), (11, 10)]),
# ]

# # Create GeoSeries
# s = gpd.GeoSeries(polygons,)

# # Calculate shortest lines between all pairs using pairwise approach
# print("Calculating shortest lines between polygons...")
# shortest_lines = s.shortest_line(s, align=False)
# print(shortest_lines)
# # # Create the plot
# fig, ax = plt.subplots(figsize=(10, 8))

# # Plot polygons with different colors
# colors = ['lightblue', 'lightgreen', 'lightcoral']
# for i, (polygon, color) in enumerate(zip(polygons, colors)):
#     x, y = polygon.exterior.xy
#     ax.fill(x, y, alpha=0.7, fc=color, ec='black', linewidth=2, label=f'Polygon {i+1}')
    
#     # Add polygon label
#     centroid = polygon.centroid
#     ax.annotate(f'P{i+1}', (centroid.x, centroid.y), 
#                fontsize=12, ha='center', va='center', weight='bold')

# # Plot shortest lines between all pairs
# line_colors = ['red', 'blue', 'green']
# line_idx = 0

# for i in range(len(polygons)):
#     for j in range(len(polygons)):
#         if i != j:  # Don't connect polygon to itself
#             try:
#                 # Create GeoSeries with proper indices to avoid alignment warning
#                 source_series = gpd.GeoSeries([s.iloc[i]], index=[0])
#                 target_series = gpd.GeoSeries([s.iloc[j]], index=[0])
                
#                 # Get the shortest line with explicit alignment
#                 shortest_lines = source_series.shortest_line(target_series, align=True)
#                 line = shortest_lines.iloc[0]
                
#                 if line and not line.is_empty:
#                     x, y = line.xy
#                     color = line_colors[line_idx % len(line_colors)]
#                     ax.plot(x, y, color=color, linewidth=2, linestyle='--', 
#                            label=f'P{i+1} to P{j+1}' if line_idx < 3 else "")
                    
#                     # Add distance label at midpoint
#                     midpoint = line.interpolate(0.5, normalized=True)
#                     distance = line.length
#                     ax.annotate(f'{distance:.2f}', (midpoint.x, midpoint.y), 
#                                fontsize=8, ha='center', va='bottom',
#                                bbox=dict(boxstyle="round,pad=0.3", facecolor='white', alpha=0.8))
#                     line_idx += 1
                    
#                     print(f"P{i+1} to P{j+1}: Length = {distance:.4f}")
#             except Exception as e:
#                 print(f"Error calculating line from P{i+1} to P{j+1}: {e}")

# # Customize the plot
# ax.set_xlim(-1, 12)
# ax.set_ylim(-1, 13)
# ax.set_xlabel('X Coordinate')
# ax.set_ylabel('Y Coordinate')
# ax.set_title('3 Separated Polygons with Shortest Connecting Lines')
# ax.grid(True, alpha=0.3)
# ax.legend()
# ax.set_aspect('equal')

# plt.tight_layout()
# plt.show()

# print("\nShortest lines calculation completed!")



"""
Functions related to  EASE-DGGS hierarchy.

© Regents of the University of Minnesota. All rights reserved.
This software is released under an Apache 2.0 license. Further details about the Apache 2.0 license are available in the license.txt file.
"""

from itertools import product

import pandas as pd

from shapely import wkt

from vgrid.utils.easedggs.constants import levels_specs
from vgrid.utils.easedggs.dggs.checks import validate_grid_ids, check_level

from vgrid.utils.easedggs.dggs.utils import format_response, enumerate_id_elements
from vgrid.utils.easedggs.dggs.utils import enumerate_grid_table_rows
from vgrid.utils.easedggs.dggs.utils import calc_grid_coord_vectors

from vgrid.utils.easedggs.logConfig import logger


def _child_to_parent(gid, level=0):
    """
    Determines the parent cell (coarser) of a child at the specified level.

    Parameters
    ----------
    gid : str
        Grid ID of the child cell.
    level : str
        The specified level (resolution) of the parent to return.

    Returns
    -------
    partent_id : str
        Parent ID of of the cell.
    """
    if not isinstance(gid, str):
        return False

    id_elements = gid.split(".")
    # test that L0 not passed in
    if id_elements[0][1] == "0":
        return False

    index_max = level + 2

    parent_id = id_elements[0:index_max]
    parent_id[0] = f"L{str(level)}"

    return ".".join(parent_id)


def children_to_parents(children, level=0):
    """
    Determines the parent cells (coarser) of all children at the specified level.

    Parameters
    ----------
    children : List
        Children whose parent cells you want to identify.
    level : str
        The level of the parent cells.

    Returns
    -------
    parent_ids : list
        Parent IDs of of the cells.
    """
    if not isinstance(children, list):
        return False

    success, data = validate_grid_ids(children)

    if not success:
        return format_response(data, success)

    data = [_child_to_parent(gid, level=level) for gid in children]

    if not any(data):
        data = ["Invalid grid ID passed to _child_to_parent "]
        return format_response(data, False)

    return format_response(data, success)


def gen_child_geometries(parent_geometry, parent_id, child_level, wkt_geom=True):
    """
    Generate child cell characterisitcs from parent cells.

    Parameters
    ----------
    parent_geometry : polygon
       WKT with parent polygone.
    parent_id : float
        The cell id of the parent cell.
    child_level _ int
        The grid level of the child to create
    wkt_geom : boolean
        Denotes that partent_geometry is wkt

    Returns
    -------
    row_ID, column_ID, Grid_ID, geoms, centroid    : tuple
        Relevant child cell characteritics: Row ID, Column ID, Grid ID, Geometry and Centroid
    """
    if wkt_geom:
        parent_geometry = wkt.loads(parent_geometry)

    min_x, min_y, max_x, max_y = parent_geometry.bounds

    x_coords, y_coords = calc_grid_coord_vectors(
        min_x=min_x,
        min_y=min_y,
        max_x=max_x,
        max_y=max_y,
        level=child_level,
        x_ascend=True,
        y_ascend=False,
    )

    r_ind, c_ind, grid_id, geoms, centroid = enumerate_grid_table_rows(
        x_coords=x_coords, y_coords=y_coords, level=child_level, parent_id=parent_id
    )

    return (r_ind, c_ind, grid_id, geoms, centroid)


def _parent_to_children(gid, level=1):
    """
    Determines all the children of a single parent the specified level.

    Parameters
    ----------
    gid : str
        EASE-DGGS ID of the parent cell.
    level : str
        Level (resolution) of the children cells.

    Returns
    -------
    children : list
        List of grid IDs for all the children of the parent at the supplied level.
    """
    if not isinstance(gid, str):
        return False

    id_elements = gid.split(".")

    parent_level = len(id_elements) - 2
    parent_cell = id_elements.pop(0)
    parent_cell = ".".join(id_elements)
    parent_cell = f"L{str(level)}.{parent_cell}"

    index_max = level

    id_elements[0:index_max]

    refine_ratio = [
        levels_specs[lv]["refine_ratio"] for lv in range(parent_level, index_max)
    ]

    level_elements = [enumerate_id_elements(rr) for rr in refine_ratio]

    level_elements = list(product(*level_elements))

    level_elements = [".".join(le) for le in level_elements]

    children = ["{}.{}".format(parent_cell, le) for le in level_elements]

    return children


def parents_to_children(grid_ids, level=1):
    """
    Determines all of the children cells of the parent cell for the specified level.

    Parameters
    ----------
    grid_id : list
        List of EASE-DGGS IDs for all the parent cells.
    level : str
        Level (resolution) of the children cells to return.

    Returns
    -------
    children : list
        List of EASE-DGGS IDs for children of the parent cells.
    """
    if not isinstance(grid_ids, list):
        data = ["Input grid IDs should be list"]
        return format_response(data, False)

    success, data = validate_grid_ids(grid_ids)
    if not success:
        return format_response(data, success)

    data = [_parent_to_children(gid, level=level) for gid in grid_ids]

    return format_response(data, success)


def grid_aggregate(
    grid_ids, grid_vals, level=0, method="mean", levels_specs=levels_specs
):
    """
    Aggregate EASE-DGGS ID to coarser spatial resolution.

    Parameters
    ----------
    grid_ids : list
       List of the EASE-DGGS IDs to aggregate.

    grid_vals : list
       List of the corresponding EASE-DGGS cell values to aggregate.

    level : str
        Target resolution grid level to aggregate to.

    method : str
        The aggregation method to employ.

    Returns
    -------
    Lists with grid_ids and aggregated values lists.
    """

    if not isinstance(grid_ids, list) and not isinstance(grid_vals, list):
        success = False
        data = "Lists expected for grid_ids and grid_vals."
        return format_response(data, success)

    if not check_level(level):
        success = False
        data = ["The specified level is invalid."]

        return format_response(data, success)

    if not all(isinstance(x, (int, float)) for x in grid_vals):
        success = False
        data = "grid_vals must be ints or floats"

        return format_response(data, success)

    allow_methods = [
        "count",
        "first",
        "last",
        "mean",
        "median",
        "min",
        "max",
        "std",
        "sum",
        "var",
        "prod",
        "mode",
    ]

    numeric_true = [
        "first",
        "last",
        "mean",
        "median",
        "min",
        "max",
        "std",
        "sum",
        "var",
        "prod",
    ]

    if method not in allow_methods:
        success = False
        data = "Invalid aggregation method supplied."
        return format_response(data, success)

    parent_ids = children_to_parents(grid_ids, level=level)
    parent_ids = parent_ids["result"]["data"]

    df = pd.DataFrame(
        {"children": grid_ids, "grid_vals": grid_vals, "parents": parent_ids}
    )

    if method in numeric_true:
        agg = df.groupby("parents").aggregate(method, numeric_only=True)
        out_ids = agg.index.to_list()
        out_vals = agg.grid_vals.to_list()
    elif method == "mode":
        agg = df.groupby("parents").grid_vals.apply(lambda x: x.mode()[0])
        out_ids = agg.index.to_list()
        out_vals = list(agg.values)
    elif method == "count":
        agg = df.groupby("parents").grid_vals.apply(lambda x: x.count())
        out_ids = agg.index.to_list()
        out_vals = list(agg.values)
    else:
        logger.warning(
            "Error in heirarchy.py. Else condition encounter that shoudl not be possible."
        )

    return format_response({"grid_ids": out_ids, "values": out_vals}, True)



import argparse
import os
import json
import statistics
from collections import defaultdict, Counter
from shapely.geometry import Point
from tqdm import tqdm
from vgrid.binning.bin_helper import get_default_stats_structure, append_stats_value
from vgrid.generator.settings import geodesic_dggs_to_feature
from vgrid.utils import qtm


def qtm_bin(point_features, resolution, stats, category, field_name):
    qtm_bins = defaultdict(lambda: defaultdict(get_default_stats_structure))

    for feature in tqdm(point_features, desc="Binning points"):
        geom = feature["geometry"]
        props = feature.get("properties", {})

        if geom["type"] == "Point":
            point = Point(geom["coordinates"])
            qtm_id = qtm.latlon_to_qtm_id(point.y, point.x, resolution)
            append_stats_value(qtm_bins, qtm_id, props, stats, category, field_name)

        elif geom["type"] == "MultiPoint":
            for coords in geom["coordinates"]:
                point = Point(coords)
                qtm_id = qtm.latlon_to_qtm_id(point.y, point.x, resolution)
                append_stats_value(qtm_bins, qtm_id, props, stats, category, field_name)

    qtm_features = []
    for qtm_id, categories in qtm_bins.items():
        facet = qtm.qtm_id_to_facet(qtm_id)
        cell_polygon = qtm.constructGeometry(facet)

        if not cell_polygon.is_valid:
            continue

        num_edges = 3
        qtm_feature = geodesic_dggs_to_feature(
            "qtm", qtm_id, resolution, cell_polygon, num_edges
        )

        for cat, values in categories.items():
            key_prefix = "" if category is None else f"{cat}_"

            if stats == "count":
                qtm_feature["properties"][f"{key_prefix}count"] = values["count"]
            elif stats == "sum":
                qtm_feature["properties"][f"{key_prefix}sum"] = sum(values["sum"])
            elif stats == "min":
                qtm_feature["properties"][f"{key_prefix}min"] = min(values["min"])
            elif stats == "max":
                qtm_feature["properties"][f"{key_prefix}max"] = max(values["max"])
            elif stats == "mean":
                qtm_feature["properties"][f"{key_prefix}mean"] = statistics.mean(
                    values["mean"]
                )
            elif stats == "median":
                qtm_feature["properties"][f"{key_prefix}median"] = statistics.median(
                    values["median"]
                )
            elif stats == "std":
                qtm_feature["properties"][f"{key_prefix}std"] = (
                    statistics.stdev(values["std"]) if len(values["std"]) > 1 else 0
                )
            elif stats == "var":
                qtm_feature["properties"][f"{key_prefix}var"] = (
                    statistics.variance(values["var"]) if len(values["var"]) > 1 else 0
                )
            elif stats == "range":
                qtm_feature["properties"][f"{key_prefix}range"] = (
                    max(values["range"]) - min(values["range"])
                    if values["range"]
                    else 0
                )

            elif stats == "minority":
                freq = Counter(values["values"])
                min_item = min(freq.items(), key=lambda x: x[1])[0] if freq else None
                qtm_feature["properties"][f"{key_prefix}minority"] = min_item

            elif stats == "majority":
                freq = Counter(values["values"])
                max_item = max(freq.items(), key=lambda x: x[1])[0] if freq else None
                qtm_feature["properties"][f"{key_prefix}majority"] = max_item

            elif stats == "variety":
                qtm_feature["properties"][f"{key_prefix}variety"] = len(
                    set(values["values"])
                )

        qtm_features.append(qtm_feature)

    return qtm_features


def main():
    parser = argparse.ArgumentParser(description="Binning point to QTM DGGS")
    parser.add_argument(
        "-point",
        "--point",
        type=str,
        required=True,
        help="GeoJSON file path (Point or MultiPoint)",
    )
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        default=14,
        help="Resolution of the grid [1..24]",
    )
    parser.add_argument(
        "-stats",
        "--statistics",
        choices=[
            "count",
            "min",
            "max",
            "sum",
            "mean",
            "median",
            "std",
            "var",
            "range",
            "minority",
            "majority",
            "variety",
        ],
        required=True,
        help="Statistic option",
    )

    parser.add_argument(
        "-category",
        "--category",
        required=False,
        help="Optional category field for grouping",
    )
    parser.add_argument(
        "-field", "--field", required=False, help="Numeric field to compute statistics"
    )

    args = parser.parse_args()

    resolution = args.resolution
    point = args.point
    stats = args.statistics
    category = args.category
    field_name = args.field

    if resolution < 1 or resolution > 24:
        print("Error: Please select a resolution in [1..24].")
        return

    if not os.path.exists(point):
        print(f"Error: The file {point} does not exist.")
        return

    if stats != "count" and not field_name:
        print("Error: A field name is required for statistics other than 'count'.")
        return

    with open(point, "r", encoding="utf-8") as f:
        point_data = json.load(f)

    point_features = point_data["features"]
    qtm_features = qtm_bin(point_features, resolution, stats, category, field_name)

    out_name = os.path.splitext(os.path.basename(point))[0]
    out_path = f"{out_name}_bin_qtm_{resolution}_{stats}.geojson"

    with open(out_path, "w", encoding="utf-8") as f:
        json.dump({"type": "FeatureCollection", "features": qtm_features}, f, indent=2)

    print(f"GeoJSON saved as {out_path}")


if __name__ == "__main__":
    main()



import sys
import argparse
import json
from tqdm import tqdm
from shapely.geometry import Polygon, shape, MultiPoint, MultiLineString, MultiPolygon
import pandas as pd
import geopandas as gpd
from vgrid.utils import olc
from vgrid.generator.olcgrid import generate_grid, refine_cell
from vgrid.generator.settings import graticule_dggs_to_feature
from vgrid.conversion.dggscompact import olccompact
from vgrid.utils.geometry import (
    check_predicate,
    shortest_point_distance,
    shortest_polyline_distance,
    shortest_polygon_distance,
)
from vgrid.stats.olcstats import olc_metrics
from math import sqrt

def validate_olc_resolution(resolution):
    """
    Validate that OLC resolution is in the valid range [2,4,6,8,10,11,12,13,14,15].

    Args:
        resolution (int): Resolution value to validate

    Returns:
        int: Validated resolution value

    Raises:
        ValueError: If resolution is not in range [2,4,6,8,10,11,12,13,14,15]
        TypeError: If resolution is not an integer
    """
    if not isinstance(resolution, int):
        raise TypeError(
            f"Resolution must be an integer, got {type(resolution).__name__}"
        )

    if resolution not in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:
        raise ValueError(
            f"Resolution must be in [2,4,6,8,10,11,12,13,14,15], got {resolution}"
        )

    return resolution


def point2olc(
    resolution,
    point,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_points=None,  # New parameter for topology preservation
):
    """
    Convert a point geometry to an OLC grid cell.

    Args:
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        point (shapely.geometry.Point): Point geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for points)
        compact (bool, optional): Enable OLC compact mode (not used for points)
        topology (bool, optional): Enable topology preserving mode - ensures disjoint points have disjoint OLC cells
        include_properties (bool, optional): Whether to include properties in output
        all_points: List of all points for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing OLC cells containing the point
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_points is None:
            raise ValueError("all_points parameter is required when topology=True")
        
        # Calculate the shortest distance between all points
        shortest_distance = shortest_point_distance(all_points)
        
        # Find resolution where OLC cell size is smaller than shortest distance
        # This ensures disjoint points have disjoint OLC cells
        if shortest_distance > 0:
            for res in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:  # OLC valid resolutions
                _, avg_edge_length, _ = olc_metrics(res)
                # Use a factor to ensure sufficient separation (cell diameter is ~2x edge length)
                cell_diameter = avg_edge_length * sqrt(2)* 2
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single point or no distance, use provided resolution
            pass

    olc_features = []
    olc_id = olc.encode(point.y, point.x, resolution)
    coord = olc.decode(olc_id)
    if coord:
        min_lat, min_lon = coord.latitudeLo, coord.longitudeLo
        max_lat, max_lon = coord.latitudeHi, coord.longitudeHi
        cell_polygon = Polygon(
            [
                [min_lon, min_lat],
                [max_lon, min_lat],
                [max_lon, max_lat],
                [min_lon, max_lat],
                [min_lon, min_lat],
            ]
        )
        olc_feature = graticule_dggs_to_feature("olc", olc_id, resolution, cell_polygon)
        if include_properties and feature_properties:
            olc_feature["properties"].update(feature_properties)
        olc_features.append(olc_feature)
    return olc_features


def polyline2olc(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polylines=None,  # New parameter for topology preservation
):
    """
    Convert line geometries (LineString, MultiLineString) to OLC grid cells.

    Args:
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        feature (shapely.geometry.LineString or shapely.geometry.MultiLineString): Line geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply (not used for lines)
        compact (bool, optional): Enable OLC compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polylines have disjoint OLC cells
        include_properties (bool, optional): Whether to include properties in output
        all_polylines: List of all polylines for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing OLC cells intersecting the line
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polylines is None:
            raise ValueError("all_polylines parameter is required when topology=True")
        
        # Calculate the shortest distance between all polylines
        shortest_distance = shortest_polyline_distance(all_polylines)
        
        # Find resolution where OLC cell size is smaller than shortest distance
        # This ensures disjoint polylines have disjoint OLC cells
        if shortest_distance > 0:
            for res in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:  # OLC valid resolutions
                _, avg_edge_length, _ = olc_metrics(res)
                # Use a factor to ensure sufficient separation (cell diameter is ~2x edge length)
                cell_diameter = avg_edge_length * sqrt(2) * 4  # in case there are 2 cells representing the same line segment
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single polyline or no distance, use provided resolution
            pass
    olc_features = []
    if feature.geom_type in ("LineString"):
        polylines = [feature]
    elif feature.geom_type in ("MultiLineString"):
        polylines = list(feature.geoms)
    else:
        return []
    for polyline in polylines:
        base_resolution = 2
        base_cells = generate_grid(base_resolution, verbose=False)
        seed_cells = []
        for base_cell in base_cells["features"]:
            base_cell_poly = Polygon(base_cell["geometry"]["coordinates"][0])
            if polyline.intersects(base_cell_poly):
                seed_cells.append(base_cell)
        refined_features = []
        for seed_cell in seed_cells:
            seed_cell_poly = Polygon(seed_cell["geometry"]["coordinates"][0])
            if seed_cell_poly.contains(polyline) and resolution == base_resolution:
                refined_features.append(seed_cell)
            else:
                refined_features.extend(
                    refine_cell(
                        seed_cell_poly.bounds, base_resolution, resolution, polyline
                    )
                )
        resolution_features = [
            refined_feature
            for refined_feature in refined_features
            if refined_feature["properties"]["resolution"] == resolution
        ]
        seen_olc_codes = set()
        for resolution_feature in resolution_features:
            olc_id = resolution_feature["properties"]["olc"]
            if olc_id not in seen_olc_codes:
                if include_properties and feature_properties:
                    resolution_feature["properties"].update(feature_properties)
                olc_features.append(resolution_feature)
                seen_olc_codes.add(olc_id)
    olc_geojson = {"type": "FeatureCollection", "features": olc_features}
    if compact:
        return olccompact(olc_geojson)["features"]
    return olc_features


def polygon2olc(
    resolution,
    feature,
    feature_properties=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
    all_polygons=None,  # New parameter for topology preservation
):
    """
    Convert polygon geometries (Polygon, MultiPolygon) to OLC grid cells.

    Args:
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        feature (shapely.geometry.Polygon or shapely.geometry.MultiPolygon): Polygon geometry to convert
        feature_properties (dict, optional): Properties to include in output features
        predicate (str, optional): Spatial predicate to apply ('intersect', 'within', 'centroid_within', 'largest_overlap')
        compact (bool, optional): Enable OLC compact mode to reduce cell count
        topology (bool, optional): Enable topology preserving mode - ensures disjoint polygons have disjoint OLC cells
        include_properties (bool, optional): Whether to include properties in output
        all_polygons: List of all polygons for topology preservation (required when topology=True)

    Returns:
        list: List of GeoJSON feature dictionaries representing OLC cells based on predicate
    """
    # If topology preservation is enabled, calculate appropriate resolution
    if topology:
        if all_polygons is None:
            raise ValueError("all_polygons parameter is required when topology=True")
        
        # Calculate the shortest distance between all polygons
        shortest_distance = shortest_polygon_distance(all_polygons)
        
        # Find resolution where OLC cell size is smaller than shortest distance
        # This ensures disjoint polygons have disjoint OLC cells
        if shortest_distance > 0:
            for res in [2, 4, 6, 8, 10, 11, 12, 13, 14, 15]:  # OLC valid resolutions
                _, avg_edge_length, _ = olc_metrics(res)
                # Use a factor to ensure sufficient separation (cell diameter is ~2x edge length)
                cell_diameter = avg_edge_length * sqrt(2) * 4
                if cell_diameter < shortest_distance:
                    resolution = res
                    break
            else:
                # If no resolution found, use the highest resolution
                resolution = 15
        else:
            # Single polygon or no distance, use provided resolution
            pass
    olc_features = []
    if feature.geom_type in ("Polygon"):
        polygons = [feature]
    elif feature.geom_type in ("MultiPolygon"):
        polygons = list(feature.geoms)
    else:
        return []
    for polygon in polygons:
        base_resolution = 2
        base_cells = generate_grid(base_resolution, verbose=False)
        seed_cells = []
        for base_cell in base_cells["features"]:
            base_cell_poly = Polygon(base_cell["geometry"]["coordinates"][0])
            if polygon.intersects(base_cell_poly):
                seed_cells.append(base_cell)
        refined_features = []
        for seed_cell in seed_cells:
            seed_cell_poly = Polygon(seed_cell["geometry"]["coordinates"][0])
            if seed_cell_poly.contains(polygon) and resolution == base_resolution:
                refined_features.append(seed_cell)
            else:
                refined_features.extend(
                    refine_cell(
                        seed_cell_poly.bounds, base_resolution, resolution, polygon
                    )
                )
        resolution_features = [
            refined_feature
            for refined_feature in refined_features
            if refined_feature["properties"]["resolution"] == resolution
        ]
        seen_olc_codes = set()
        for resolution_feature in resolution_features:
            olc_id = resolution_feature["properties"]["olc"]
            if olc_id not in seen_olc_codes:
                cell_geom = shape(resolution_feature["geometry"])
                if not check_predicate(cell_geom, polygon, predicate):
                    continue
                if include_properties and feature_properties:
                    resolution_feature["properties"].update(feature_properties)
                olc_features.append(resolution_feature)
                seen_olc_codes.add(olc_id)
    olc_geojson = {"type": "FeatureCollection", "features": olc_features}
    if compact:
        return olccompact(olc_geojson)["features"]
    return olc_features


def geometry2olc(
    geometries,
    resolution,
    properties_list=None,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a list of geometries to OLC grid cells.

    Args:
        geometries (shapely.geometry.BaseGeometry or list): Single geometry or list of geometries
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        properties_list (list, optional): List of property dictionaries for each geometry
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable OLC compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with OLC grid cells
    """
    resolution = validate_olc_resolution(resolution)

    # Handle single geometry or list of geometries
    if not isinstance(geometries, list):
        geometries = [geometries]

    # Handle properties
    if properties_list is None:
        properties_list = [{} for _ in geometries]
    elif not isinstance(properties_list, list):
        properties_list = [properties_list for _ in geometries]

    # Collect all points, polylines, and polygons for topology preservation if needed
    all_points = None
    all_polylines = None
    all_polygons = None
    if topology:
        points_list = []
        polylines_list = []
        polygons_list = []
        for i, geom in enumerate(geometries):
            if geom is None:
                continue
            if geom.geom_type == "Point":
                points_list.append(geom)
            elif geom.geom_type == "MultiPoint":
                points_list.extend(list(geom.geoms))
            elif geom.geom_type == "LineString":
                polylines_list.append(geom)
            elif geom.geom_type == "MultiLineString":
                polylines_list.extend(list(geom.geoms))
            elif geom.geom_type == "Polygon":
                polygons_list.append(geom)
            elif geom.geom_type == "MultiPolygon":
                polygons_list.extend(list(geom.geoms))
        if points_list:
            all_points = MultiPoint(points_list)
        if polylines_list:
            all_polylines = MultiLineString(polylines_list)
        if polygons_list:
            all_polygons = MultiPolygon(polygons_list)

    olc_features = []

    for i, geom in enumerate(tqdm(geometries, desc="Processing features")):
        if geom is None:
            continue

        # Get properties for this geometry
        props = properties_list[i] if i < len(properties_list) else {}

        # Process based on geometry type
        if geom.geom_type == "Point":
            point_features = point2olc(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_points,  # Pass all points for topology preservation
            )
            olc_features.extend(point_features)

        elif geom.geom_type == "MultiPoint":
            for point in geom.geoms:
                point_features = point2olc(
                    resolution,
                    point,
                    props,
                    predicate,
                    compact,
                    topology,
                    include_properties,
                    all_points,  # Pass all points for topology preservation
                )
                olc_features.extend(point_features)

        elif geom.geom_type in ["LineString", "MultiLineString"]:
            polyline_features = polyline2olc(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_polylines,  # Pass all polylines for topology preservation
            )
            olc_features.extend(polyline_features)

        elif geom.geom_type in ["Polygon", "MultiPolygon"]:
            poly_features = polygon2olc(
                resolution,
                geom,
                props,
                predicate,
                compact,
                topology,
                include_properties,
                all_polygons=all_polygons,  # Pass all polygons for topology preservation
            )
            olc_features.extend(poly_features)

        else:
            raise ValueError(f"Unsupported geometry type: {geom.geom_type}")

    return {"type": "FeatureCollection", "features": olc_features}


def dataframe2olc(
    df,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a pandas DataFrame with geometry column to OLC grid cells.

    Args:
        df (pandas.DataFrame): DataFrame with geometry column
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable OLC compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with OLC grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in df.iterrows():
        geom = row.geometry if "geometry" in row else row["geometry"]
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2olc(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def geodataframe2olc(
    gdf,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    include_properties=True,
):
    """
    Convert a GeoDataFrame to OLC grid cells.

    Args:
        gdf (geopandas.GeoDataFrame): GeoDataFrame to convert
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        predicate (str, optional): Spatial predicate to apply for polygons
        compact (bool, optional): Enable OLC compact mode for polygons and lines
        topology (bool, optional): Enable topology preserving mode
        include_properties (bool, optional): Whether to include properties in output

    Returns:
        dict: GeoJSON FeatureCollection with OLC grid cells
    """
    geometries = []
    properties_list = []
    for idx, row in gdf.iterrows():
        geom = row.geometry
        if geom is not None:
            geometries.append(geom)
            props = row.to_dict()
            if "geometry" in props:
                del props["geometry"]
            properties_list.append(props)
    return geometry2olc(
        geometries,
        resolution,
        properties_list,
        predicate,
        compact,
        topology,
        include_properties,
    )


def vector2olc(
    data,
    resolution,
    predicate=None,
    compact=False,
    topology=False,
    output_format="geojson",
    output_path=None,
    include_properties=True,
    **kwargs,
):
    """
    Convert vector data to OLC grid cells from various input formats.

    Args:
        data: File path, URL, DataFrame, GeoJSON dict, or Shapely geometry
        resolution (int): OLC resolution [2,4,6,8,10,11,12,13,14,15]
        compact (bool): Enable OLC compact mode for polygons (default: False)
        output_format (str): Output format ('geojson', 'gpkg', 'parquet', 'csv', 'shapefile')
        output_path (str): Output file path (optional)
        include_properties (bool): If False, do not include original feature properties. (default: True)
        **kwargs: Additional arguments passed to geopandas read functions
    Returns:
        dict or str: Output in the specified format
    """
    if hasattr(data, "geometry") and hasattr(data, "columns"):
        result = geodataframe2olc(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif isinstance(data, pd.DataFrame):
        result = dataframe2olc(
            data, resolution, predicate, compact, topology, include_properties
        )
    elif hasattr(data, "geom_type") or (
        isinstance(data, list) and len(data) > 0 and hasattr(data[0], "geom_type")
    ):
        result = geometry2olc(
            data, resolution, None, predicate, compact, topology, include_properties
        )
    elif isinstance(data, dict) and "type" in data:
        try:
            gdf = gpd.GeoDataFrame.from_features(data["features"])
            result = geodataframe2olc(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to convert GeoJSON to GeoDataFrame: {str(e)}")
    elif isinstance(data, str):
        try:
            gdf = gpd.read_file(data, **kwargs)
            result = geodataframe2olc(
                gdf, resolution, predicate, compact, topology, include_properties
            )
        except Exception as e:
            raise ValueError(f"Failed to read file/URL {data}: {str(e)}")
    else:
        raise ValueError(f"Unsupported input type: {type(data)}")
    return convert_to_output_format(result, output_format, output_path)


def convert_to_output_format(result, output_format, output_path=None):
    """
    Convert OLC result to specified output format.

    Args:
        result (dict): GeoJSON FeatureCollection result
        output_format (str): Desired output format
        output_path (str): Output file path (optional)

    Returns:
        dict or str: Output in the specified format
    """
    # Check if result has features
    if not result or "features" not in result or not result["features"]:
        print("Warning: No features found in result. This may happen when:")
        print("  - Using 'within' predicate with coarse resolution (cells too large)")
        print("  - Using 'largest_overlap' predicate with no cells having >50% overlap")
        print("  - Input geometry is invalid or empty")
        print("Suggestions:")
        print("  - Try a finer resolution (higher number)")
        print("  - Use 'intersect' or 'centroid_within' predicate instead")
        print("  - Check that input geometry is valid")
        raise ValueError("No features found in result")
    
    # First convert GeoJSON result to GeoDataFrame
    try:
        gdf = gpd.GeoDataFrame.from_features(result["features"])

        # Set CRS to WGS84 (EPSG:4326) since OLC uses WGS84 coordinates
        gdf.set_crs(epsg=4326, inplace=True)
        
        # Ensure the geometry column is set as the active geometry column
        if 'geometry' in gdf.columns:
            gdf.set_geometry('geometry', inplace=True)
        else:
            # If no geometry column found, try to identify it
            geom_cols = [col for col in gdf.columns if hasattr(gdf[col].iloc[0], 'geom_type')]
            if geom_cols:
                gdf.set_geometry(geom_cols[0], inplace=True)
            else:
                raise ValueError("No geometry column found in GeoDataFrame")
        
        # Verify the GeoDataFrame has valid geometry
        if gdf.empty:
            raise ValueError("GeoDataFrame is empty")
        
        if not gdf.geometry.is_valid.all():
            print("Warning: Some geometries are invalid")
    
    except Exception as e:
        print(f"Error creating GeoDataFrame: {str(e)}")
        print(f"Result features count: {len(result['features']) if 'features' in result else 0}")
        if 'features' in result and result['features']:
            print(f"First feature: {result['features'][0]}")
        raise

    if output_format.lower() == "geojson":
        if output_path:
            import json

            with open(output_path, "w", encoding="utf-8") as f:
                json.dump(result, f)
            return output_path
        else:
            return result  # Already in GeoJSON format

    elif output_format.lower() == "gpkg":
        if output_path:
            gdf.to_file(output_path, driver="GPKG")
            return output_path
        else:
            gdf.to_file("vector2olc.gpkg", driver="GPKG")
            return "vector2olc.gpkg"

    elif output_format.lower() == "parquet":
        if output_path:
            gdf.to_parquet(output_path, index=False)
            return output_path
        else:
            gdf.to_parquet("vector2olc.parquet", index=False)
            return "vector2olc.parquet"

    elif output_format.lower() == "csv":
        if output_path:
            gdf.to_csv(output_path, index=False)
            return output_path
        else:
            return gdf.to_csv(index=False)

    elif output_format.lower() == "shapefile":
        if output_path:
            gdf.to_file(output_path, driver="ESRI Shapefile")
            return output_path
        else:
            gdf.to_file("vector2olc.shp", driver="ESRI Shapefile")
            return "vector2olc.shp"

    else:
        raise ValueError(
            f"Unsupported output format: {output_format}. Supported formats: geojson, gpkg, parquet, csv, shapefile"
        )


def vector2olc_cli():
    """
    Command-line interface for vector2olc conversion.

    Usage:
        python vector2olc.py -i input.shp -r 10 -c -f geojson -o output.geojson

    Arguments:
        -i, --input: Input file path or URL
        -r, --resolution: OLC resolution (see OLC spec)
        -c, --compact: Enable OLC compact mode
        -p, --predicate: Spatial predicate (intersect, within, centroid_within, largest_overlap)
        -t, --topology: Enable topology preserving mode
        -np, --no-props: Do not include original feature properties
        -f, --format: Output format (geojson, gpkg, parquet, csv, shapefile)
        -o, --output: Output file path
    """
    parser = argparse.ArgumentParser(
        description="Convert vector data to OLC grid cells"
    )
    parser.add_argument("-i", "--input", help="Input file path, URL")
    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        choices=[2, 4, 6, 8, 10, 11, 12, 13, 14, 15],
        metavar="[2,4,6,8,10,11,12,13,14,15]",
        help="OLC resolution (see OLC spec)",
    )
    parser.add_argument(
        "-c",
        "--compact",
        action="store_true",
        help="Enable OLC compact mode for polygons",
    )
    parser.add_argument(
        "-p",
        "--predicate",
        choices=["intersect", "within", "centroid_within", "largest_overlap"],
        help="Spatial predicate: intersect, within, centroid_within, largest_overlap for polygons",
    )
    parser.add_argument(
        "-t", "--topology", action="store_true", 
        help="Enable topology preserving mode ensuring disjoint features have disjoint OLC cells by automatically calculating appropriate resolution."
    )
    parser.add_argument(
        "-np",
        "-no-props",
        dest="include_properties",
        action="store_false",
        help="Do not include original feature properties.",
    )
    parser.add_argument(
        "-f",
        "--format",
        default="geojson",
        choices=["geojson", "gpkg", "parquet", "csv", "shapefile"],
        help="Output format (default: geojson)",
    )
    parser.add_argument("-o", "--output", help="Output file path (optional)")
    args = parser.parse_args()
    if args.resolution is not None:
        try:
            args.resolution = validate_olc_resolution(args.resolution)
        except (ValueError, TypeError) as e:
            print(f"Error: {e}", file=sys.stderr)
            sys.exit(1)
    data = args.input
    output_path = args.output
    if not output_path and args.format in [
        "geojson",
        "gpkg",
        "parquet",
        "csv",
        "shapefile",
    ]:
        extensions = {
            "geojson": ".geojson",
            "gpkg": ".gpkg",
            "parquet": ".parquet",
            "csv": ".csv",
            "shapefile": ".shp",
        }
        output_path = f"vector2olc{extensions.get(args.format, '')}"
    try:
        vector2olc(
            data,
            args.resolution,
            predicate=args.predicate,
            compact=args.compact,
            topology=args.topology,
            output_format=args.format,
            output_path=output_path,
            include_properties=args.include_properties,
        )
        if output_path:
            print(f"Output saved to {output_path}")
    except Exception as e:
        print(f"Error: {str(e)}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    vector2olc_cli()



import os
import argparse
import json
from tqdm import tqdm
import rasterio
import h3
import numpy as np
from shapely.geometry import Polygon
import csv
from vgrid.generator.h3grid import fix_h3_antimeridian_cells
from vgrid.generator.settings import geodesic_dggs_to_feature
from math import cos, radians


def get_nearest_h3_resolution(raster_path):
    with rasterio.open(raster_path) as src:
        transform = src.transform
        crs = src.crs
        pixel_width = transform.a
        pixel_height = -transform.e
        cell_size = pixel_width * pixel_height

        if crs.is_geographic:
            # Latitude of the raster center
            center_latitude = (src.bounds.top + src.bounds.bottom) / 2
            # Convert degrees to meters
            meter_per_degree_lat = 111_320  # Roughly 1 degree latitude in meters
            meter_per_degree_lon = meter_per_degree_lat * cos(radians(center_latitude))

            pixel_width_m = pixel_width * meter_per_degree_lon
            pixel_height_m = pixel_height * meter_per_degree_lat
            cell_size = pixel_width_m * pixel_height_m

    nearest_resolution = None
    min_diff = float("inf")

    # Check resolutions from 0 to 15
    for res in range(16):
        avg_area = h3.average_hexagon_area(res, unit="m^2")
        diff = abs(avg_area - cell_size)
        # If the difference is smaller than the current minimum, update the nearest resolution
        if diff < min_diff:
            min_diff = diff
            nearest_resolution = res

    return nearest_resolution


def convert_numpy_types(obj):
    """Recursively convert NumPy types to native Python types"""
    if isinstance(obj, np.generic):
        return obj.item()  # Convert numpy types like np.uint8 to native Python int
    elif isinstance(obj, dict):
        return {k: convert_numpy_types(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return [convert_numpy_types(v) for v in obj]
    else:
        return obj


def raster2h3(raster_path, resolution=None, format="geojson"):
    # Step 1: Determine the nearest H3 resolution if none is provided
    if resolution is None:
        resolution = get_nearest_h3_resolution(raster_path)
        print(f"Nearest H3 resolution determined: {resolution}")

    # Validate resolution is in valid range
    if resolution < 0 or resolution > 15:
        raise ValueError("Resolution must be in range [0..15]")

    # Open the raster file to get metadata and data
    with rasterio.open(raster_path) as src:
        raster_data = src.read()  # Read all bands
        transform = src.transform
        width, height = src.width, src.height
        band_count = src.count  # Number of bands in the raster

    h3_ids = set()

    for row in range(height):
        for col in range(width):
            lon, lat = transform * (col, row)
            h3_id = h3.latlng_to_cell(lat, lon, resolution)
            h3_ids.add(h3_id)

    # Sample the raster values at the centroids of the H3 hexagons
    h3_data = []

    for h3_id in tqdm(h3_ids, desc="Resampling", unit=" cells"):
        # Get the centroid of the H3 cell
        centroid_lat, centroid_lon = h3.cell_to_latlng(h3_id)
        # Sample the raster values at the centroid (lat, lon)
        col, row = ~transform * (centroid_lon, centroid_lat)

        if 0 <= col < width and 0 <= row < height:
            # Get the values for all bands at this centroid
            values = raster_data[:, int(row), int(col)]
            h3_data.append(
                {
                    "h3": h3_id,
                    **{
                        f"band_{i + 1}": values[i] for i in range(band_count)
                    },  # Create separate columns for each band
                }
            )

    if format.lower() == "csv":
        import io

        output = io.StringIO()
        if h3_data:
            writer = csv.DictWriter(output, fieldnames=h3_data[0].keys())
            writer.writeheader()
            writer.writerows(h3_data)
        return output.getvalue()

    # Create the GeoJSON-like structure
    h3_features = []
    for data in tqdm(h3_data, desc="Converting to GeoJSON", unit=" cells"):
        cell_boundary = h3.cell_to_boundary(data["h3"])
        if cell_boundary:
            filtered_boundary = fix_h3_antimeridian_cells(cell_boundary)
            # Reverse lat/lon to lon/lat for GeoJSON compatibility
            reversed_boundary = [(lon, lat) for lat, lon in filtered_boundary]
            cell_polygon = Polygon(reversed_boundary)
            num_edges = 6
            if h3.is_pentagon(data["h3"]):
                num_edges = 5

            h3_feature = geodesic_dggs_to_feature(
                "h3", data["h3"], resolution, cell_polygon, num_edges
            )

            band_properties = {
                f"band_{i + 1}": data[f"band_{i + 1}"] for i in range(band_count)
            }
            h3_feature["properties"].update(convert_numpy_types(band_properties))
            h3_features.append(h3_feature)

    return {
        "type": "FeatureCollection",
        "features": h3_features,
    }


def raster2h3_cli():
    parser = argparse.ArgumentParser(
        description="Convert Raster in Geographic CRS to H3 DGGS"
    )
    parser.add_argument("-raster", type=str, required=True, help="Raster file path")

    parser.add_argument(
        "-r",
        "--resolution",
        type=int,
        required=False,
        default=None,
        help="Resolution [0..15]",
        min=0,
        max=15,
    )

    parser.add_argument(
        "-f",
        "--format",
        type=str,
        required=False,
        default="geojson",
        choices=["geojson", "csv"],
        help="Output format (geojson or csv)",
    )

    args = parser.parse_args()
    raster = args.raster
    resolution = args.resolution
    format = args.format

    if not os.path.exists(raster):
        print(f"Error: The file {raster} does not exist.")
        return
    if resolution is not None:
        if resolution < 0 or resolution > 15:
            print("Please select a resolution in [0..15] range and try again ")
            return

    result = raster2h3(raster, resolution, format)
    base_name = os.path.splitext(os.path.basename(raster))[0]
    output_path = f"{base_name}2h3.{format}"

    if format.lower() == "geojson":
        with open(output_path, "w") as f:
            json.dump(result, f)

    elif format.lower() == "csv":
        # Get all possible field names from the first row
        fieldnames = list(result[0].keys())
        with open(output_path, "w", newline="") as f:
            writer = csv.DictWriter(f, fieldnames=fieldnames)
            writer.writeheader()
            writer.writerows(result)

    print(f"Output saved as {output_path}")



from vgrid.utils import georef
import json
import argparse
from tqdm import tqdm
from shapely.geometry import Polygon, box
import numpy as np

from vgrid.generator.settings import max_cells, graticule_dggs_to_feature

RESOLUTION_DEGREES = {
    -1: 15.0,  # 15° x 15°
    0: 1.0,  # 1° x 1°
    1: 1 / 60,  # 1-minute
    2: 1 / 600,  # 0.1-minute
    3: 1 / 6000,  # 0.01-minute
    4: 1 / 60_000,  # 0.001-minute
    5: 1 / 600_000,  # 0.0001-minute
}

# RESOLUTION_DEGREES = {
#     0: 15.0,       # 15° x 15°
#     1: 1.0,        # 1° x 1°
#     2: 1 / 60,     # 1-minute
#     3: 1 / 600,    # 0.1-minute
#     5: 1 / 6000,   # 0.01-minute
#     5: 1 / 60_000,  # 0.001-minute
#     # 5: 1 / 600_000  # 0.0001-minute
# }


def generate_grid(bbox, resolution):
    lon_min, lat_min, lon_max, lat_max = bbox
    resolution_degrees = RESOLUTION_DEGREES[resolution]
    longitudes = np.arange(lon_min, lon_max, resolution_degrees)
    latitudes = np.arange(lat_min, lat_max, resolution_degrees)
    num_cells = len(longitudes) * len(latitudes)

    print(f"Resolution {resolution} will generate {num_cells} cells ")
    if num_cells > max_cells:
        print(f"which exceeds the limit of {max_cells}.")
        print("Please select a smaller resolution and try again.")
        return
    georef_features = []

    with tqdm(total=num_cells, desc="Generating GEOREF DGGS", unit=" cells") as pbar:
        for lon in longitudes:
            for lat in latitudes:
                cell_polygon = Polygon(
                    box(lon, lat, lon + resolution_degrees, lat + resolution_degrees)
                )
                georef_id = georef.encode(lat, lon, resolution)
                georef_feature = graticule_dggs_to_feature(
                    "georef", georef_id, resolution, cell_polygon
                )
                georef_features.append(georef_feature)
                pbar.update(1)

    return {
        "type": "FeatureCollection",
        "features": georef_features,
    }


def main():
    parser = argparse.ArgumentParser(description="Generate GEOREF DGGS")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="Resolution [-1..5]"
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )

    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox if args.bbox else [-180, -90, 180, 90]

    if resolution < -1 or resolution > 5:
        print("Please select a resolution in [-1..5] range and try again ")
        return

    geojson_features = generate_grid(bbox, resolution)
    output_filename = f"georef_grid_{resolution}.geojson"

    with open(output_filename, "w") as f:
        json.dump(geojson_features, f, indent=2)

    print(f"GEOREF grid saved to {output_filename}")


if __name__ == "__main__":
    main()



"""
Copyright (c) Riskaware 2015

This file is part of OpenEAGGR.

OpenEAGGR is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

OpenEAGGR is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

A copy of the GNU Lesser General Public License is available in COPYING.LESSER
or can be found at <http://www.gnu.org/licenses/>.
"""

import unittest
from vgrid.utils.eaggr.enums.dggs_shape_location import DggsShapeLocation
from vgrid.utils.eaggr.enums.dggs_shape_type import DggsShapeType
from vgrid.utils.eaggr.shapes.dggs_cell import DggsCell
from vgrid.utils.eaggr.shapes.dggs_linestring import DggsLinestring
from vgrid.utils.eaggr.shapes.dggs_polygon import DggsPolygon
from vgrid.utils.eaggr.shapes.dggs_shape import DggsShape


## Unit tests for the DggsShape class.
class TestDggsShape(unittest.TestCase):
    def test_getters(self):
        orig_cells = [DggsCell("0000"), DggsCell("1111")]
        # Cell
        shape = DggsShape(orig_cells[0], DggsShapeLocation.ONE_FACE)
        self.assertEqual(shape.get_shape_type(), DggsShapeType.CELL)
        self.assertEqual(shape.get_shape(), orig_cells[0])
        self.assertEqual(shape.get_location(), DggsShapeLocation.ONE_FACE)
        # Linestring
        linestring = DggsLinestring()
        linestring.add_cell(orig_cells[1])
        shape = DggsShape(linestring, DggsShapeLocation.NO_LOCATION)
        self.assertEqual(shape.get_shape_type(), DggsShapeType.LINESTRING)
        linestring_cells = shape.get_shape().get_cells()
        self.assertEqual(len(linestring_cells), 1)
        self.assertEqual(linestring_cells[0], orig_cells[1])
        self.assertEqual(shape.get_location(), DggsShapeLocation.NO_LOCATION)
        # Polygon
        shape = DggsShape(DggsPolygon(linestring), DggsShapeLocation.NO_LOCATION)
        self.assertEqual(shape.get_shape_type(), DggsShapeType.POLYGON)
        outer_ring_cells = shape.get_shape().get_outer_ring().get_cells()
        self.assertEqual(len(outer_ring_cells), 1)
        self.assertEqual(outer_ring_cells[0], orig_cells[1])
        self.assertEqual(shape.get_location(), DggsShapeLocation.NO_LOCATION)

    def test_errors(self):
        # Invalid shape
        with self.assertRaises(ValueError):
            DggsShape("Invalid shape", DggsShapeLocation.NO_LOCATION)
        # Invalid location type
        with self.assertRaises(ValueError):
            DggsShape(DggsCell("0000"), "Not a DggsShapeLocation")



# Reference: https://geohash.softeng.co/uekkn, https://github.com/vinsci/geohash, https://www.movable-type.co.uk/scripts/geohash.html?geohash=dp3
import vgrid.utils.geohash as geohash
import argparse
import json
from shapely.geometry import Polygon, shape
from shapely.ops import unary_union
from tqdm import tqdm
from vgrid.generator.settings import max_cells, graticule_dggs_to_feature

initial_geohashes = [
    "b",
    "c",
    "f",
    "g",
    "u",
    "v",
    "y",
    "z",
    "8",
    "9",
    "d",
    "e",
    "s",
    "t",
    "w",
    "x",
    "0",
    "1",
    "2",
    "3",
    "p",
    "q",
    "r",
    "k",
    "m",
    "n",
    "h",
    "j",
    "4",
    "5",
    "6",
    "7",
]


def geohash_to_polygon(gh):
    """Convert geohash to a Shapely Polygon."""
    lat, lon = geohash.decode(gh)
    lat_err, lon_err = geohash.decode_exactly(gh)[2:]

    bbox = {
        "w": max(lon - lon_err, -180),
        "e": min(lon + lon_err, 180),
        "s": max(lat - lat_err, -85.051129),
        "n": min(lat + lat_err, 85.051129),
    }

    return Polygon(
        [
            (bbox["w"], bbox["s"]),
            (bbox["w"], bbox["n"]),
            (bbox["e"], bbox["n"]),
            (bbox["e"], bbox["s"]),
            (bbox["w"], bbox["s"]),
        ]
    )


def expand_geohash(gh, target_length, geohashes):
    if len(gh) == target_length:
        geohashes.add(gh)
        return
    for char in "0123456789bcdefghjkmnpqrstuvwxyz":
        expand_geohash(gh + char, target_length, geohashes)


def generate_grid(resolution):
    """Generate GeoJSON for the entire world at the given geohash resolution."""

    geohashes = set()
    for gh in initial_geohashes:
        expand_geohash(gh, resolution, geohashes)

    geohash_features = []
    for gh in tqdm(geohashes, desc="Generating Geohash DGGS", unit=" cells"):
        cell_polygon = geohash_to_polygon(gh)
        geohash_feature = graticule_dggs_to_feature(
            "geohash", gh, resolution, cell_polygon
        )
        geohash_features.append(geohash_feature)

    return {"type": "FeatureCollection", "features": geohash_features}


def expand_geohash_bbox(gh, target_length, geohashes, bbox_polygon):
    """Expand geohash only if it intersects the bounding box."""
    polygon = geohash_to_polygon(gh)
    if not polygon.intersects(bbox_polygon):
        return

    if len(gh) == target_length:
        geohashes.add(gh)  # Add to the set if it reaches the target resolution
        return

    for char in "0123456789bcdefghjkmnpqrstuvwxyz":
        expand_geohash_bbox(gh + char, target_length, geohashes, bbox_polygon)


def generate_grid_within_bbox(resolution, bbox):
    """Generate GeoJSON for geohashes within a bounding box at the given resolution."""
    geohash_features = []
    bbox_polygon = Polygon.from_bounds(*bbox)

    # Compute intersected geohashes using set comprehension
    intersected_geohashes = {
        gh
        for gh in initial_geohashes
        if geohash_to_polygon(gh).intersects(bbox_polygon)
    }

    # Expand geohash bounding box
    geohashes_bbox = set()
    for gh in intersected_geohashes:
        expand_geohash_bbox(gh, resolution, geohashes_bbox, bbox_polygon)

    # Generate GeoJSON features
    geohash_features.extend(
        graticule_dggs_to_feature("geohash", gh, resolution, geohash_to_polygon(gh))
        for gh in tqdm(geohashes_bbox, desc="Generating Geohash DGGS", unit=" cells")
    )

    return {"type": "FeatureCollection", "features": geohash_features}


def generate_grid_resample(resolution, geojson_features):
    """Generate GeoJSON for geohashes within a GeoJSON feature collection at the given resolution."""
    geohash_features = []

    # Union of all input geometries
    geometries = [
        shape(feature["geometry"]) for feature in geojson_features["features"]
    ]
    unified_geom = unary_union(geometries)

    # Compute intersected geohashes from the initial set
    intersected_geohashes = {
        gh
        for gh in initial_geohashes
        if geohash_to_polygon(gh).intersects(unified_geom)
    }

    # Expand geohash coverage within the unified geometry
    geohashes_geom = set()
    for gh in intersected_geohashes:
        expand_geohash_bbox(gh, resolution, geohashes_geom, unified_geom)

    # Generate GeoJSON features
    geohash_features.extend(
        graticule_dggs_to_feature("geohash", gh, resolution, geohash_to_polygon(gh))
        for gh in tqdm(geohashes_geom, desc="Generating Geohash DGGS", unit="cells")
    )

    return {"type": "FeatureCollection", "features": geohash_features}


def main():
    parser = argparse.ArgumentParser(description="Generate Geohash DGGS.")
    parser.add_argument(
        "-r", "--resolution", type=int, required=True, help="Resolution [1..10]"
    )
    parser.add_argument(
        "-b",
        "--bbox",
        type=float,
        nargs=4,
        help="Bounding box in the format: min_lon min_lat max_lon max_lat (default is the whole world)",
    )
    args = parser.parse_args()
    resolution = args.resolution
    bbox = args.bbox

    if not (1 <= resolution <= 10):
        print("Resolution must be between 1 and 10.")
        return

    # Validate resolution and calculate metrics
    if not bbox:
        total_cells = 32**resolution
        print(f"Resolution {resolution} will generate {total_cells} cells ")
        if total_cells > max_cells:
            print(f"which exceeds the limit of {max_cells} cells.")
            print("Please select a smaller resolution and try again.")
            return

        geojson_features = generate_grid(resolution)

    else:
        # Generate grid within the bounding box
        geojson_features = generate_grid_within_bbox(resolution, bbox)

    if geojson_features:
        # Define the GeoJSON file path
        geojson_path = f"geohash_grid_{resolution}.geojson"
        with open(geojson_path, "w") as f:
            json.dump(geojson_features, f, indent=2)

        print(f"GeoJSON saved as {geojson_path}")


if __name__ == "__main__":
    main()



import json
import re
import os
import argparse


def fix_content(input_file, output_file):
    with open(input_file, "r") as infile:
        content = infile.read()
    # Ensure all property names are in double quotes
    # content = re.sub(r'(\w+):', r'"\1":', content)
    # Ensure 'name' properties are wrapped in double quotes for values that are numbers or hexadecimal strings
    content = re.sub(r'("name":)([^\s,}]+)', r'\1"\2"', content)

    try:
        data = json.loads(content)
    except json.JSONDecodeError as e:
        print(f"Error parsing corrected GeoJSON: {e}")
        return

    with open(output_file, "w") as outfile:
        json.dump(data, outfile, indent=2)

    print(f"GeoJSON has been fixed and saved to {output_file}")


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(description="Fix GeoJSON content from DGGRID")
    parser.add_argument("input", help="Input GeoJSON file")

    # Parse the arguments
    args = parser.parse_args()

    # Derive output filename by appending "_fixed" to the input filename
    # output_file = f"{args.input.rsplit('.', 1)[0]}_content_fixed.geojson"
    output_file = os.path.join(
        os.getcwd(),
        f"{os.path.splitext(os.path.basename(args.input))[0]}_content_fixed.geojson",
    )

    # Call the function with the input and output file
    fix_content(args.input, output_file)


if __name__ == "__main__":
    main()
